#!/usr/bin/python
# Copyright (c) 2018, Evert Mulder <evertmulder@gmail.com> (base on manageiq_user.py by Daniel Korn <korndaniel1@gmail.com>)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: manageiq_group

short_description: Management of groups in ManageIQ
extends_documentation_fragment:
  - community.general._manageiq
  - community.general._attributes

author: Evert Mulder (@evertmulder)
description:
  - The manageiq_group module supports adding, updating and deleting groups in ManageIQ.
requirements:
  - manageiq-client

attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  state:
    type: str
    description:
      - V(absent) - group should not exist,
      - V(present) - group should exist.
    choices: ['absent', 'present']
    default: 'present'
  description:
    type: str
    description:
      - The group description.
    required: true
  role_id:
    type: int
    description:
      - The the group role ID.
  role:
    type: str
    description:
      - The the group role name.
      - The O(role_id) has precedence over the O(role) when supplied.
    default:
  tenant_id:
    type: int
    description:
      - The tenant for the group identified by the tenant ID.
    default:
  tenant:
    type: str
    description:
      - The tenant for the group identified by the tenant name.
      - The O(tenant_id) has precedence over the O(tenant) when supplied.
      - Tenant names are case sensitive.
    default:
  managed_filters:
    description: The tag values per category.
    type: dict
    default:
  managed_filters_merge_mode:
    type: str
    description:
      - In V(merge) mode existing categories are kept or updated, new categories are added.
      - In V(replace) mode all categories are replaced with the supplied O(managed_filters).
    choices: [merge, replace]
    default: replace
  belongsto_filters:
    description: A list of strings with a reference to the allowed host, cluster or folder.
    type: list
    elements: str
    default:
  belongsto_filters_merge_mode:
    type: str
    description:
      - In merge mode existing settings are merged with the supplied O(belongsto_filters).
      - In replace mode current values are replaced with the supplied O(belongsto_filters).
    choices: [merge, replace]
    default: replace
"""

EXAMPLES = r"""
- name: Create a group in ManageIQ with the role EvmRole-user and tenant 'my_tenant'
  community.general.manageiq_group:
    description: 'MyGroup-user'
    role: 'EvmRole-user'
    tenant: 'my_tenant'
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when you trust the network!

- name: Create a group in ManageIQ with the role EvmRole-user and tenant with tenant_id 4
  community.general.manageiq_group:
    description: 'MyGroup-user'
    role: 'EvmRole-user'
    tenant_id: 4
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when you trust the network!

- name:
    - Create or update a group in ManageIQ with the role EvmRole-user and tenant my_tenant.
    - Apply 3 prov_max_cpu and 2 department tags to the group.
    - Limit access to a cluster for the group.
  community.general.manageiq_group:
    description: 'MyGroup-user'
    role: 'EvmRole-user'
    tenant: my_tenant
    managed_filters:
      prov_max_cpu:
        - '1'
        - '2'
        - '4'
      department:
        - defense
        - engineering
    managed_filters_merge_mode: replace
    belongsto_filters:
      - "/belongsto/ExtManagementSystem|ProviderName/EmsFolder|Datacenters/EmsFolder|dc_name/EmsFolder|host/EmsCluster|Cluster name"
    belongsto_filters_merge_mode: merge
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'
      validate_certs: false # only do this when you trust the network!

- name: Delete a group in ManageIQ
  community.general.manageiq_group:
    state: 'absent'
    description: 'MyGroup-user'
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      username: 'admin'
      password: 'smartvm'

- name: Delete a group in ManageIQ using a token
  community.general.manageiq_group:
    state: 'absent'
    description: 'MyGroup-user'
    manageiq_connection:
      url: 'http://127.0.0.1:3000'
      token: 'sometoken'
"""

RETURN = r"""
group:
  description: The group.
  returned: success
  type: complex
  contains:
    description:
      description: The group description.
      returned: success
      type: str
    id:
      description: The group ID.
      returned: success
      type: int
    group_type:
      description: The group type, system or user.
      returned: success
      type: str
    role:
      description: The group role name.
      returned: success
      type: str
    tenant:
      description: The group tenant name.
      returned: success
      type: str
    managed_filters:
      description: The tag values per category.
      returned: success
      type: dict
    belongsto_filters:
      description: A list of strings with a reference to the allowed host, cluster or folder.
      returned: success
      type: list
    created_on:
      description: Group creation date.
      returned: success
      type: str
      sample: "2018-08-12T08:37:55+00:00"
    updated_on:
      description: Group update date.
      returned: success
      type: int
      sample: "2018-08-12T08:37:55+00:00"
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._manageiq import ManageIQ, manageiq_argument_spec


class ManageIQgroup:
    """
    Object to execute group management operations in manageiq.
    """

    def __init__(self, manageiq):
        self.manageiq = manageiq

        self.module = self.manageiq.module
        self.api_url = self.manageiq.api_url
        self.client = self.manageiq.client

    def group(self, description):
        """Search for group object by description.
        Returns:
            the group, or None if group was not found.
        """
        groups = self.client.collections.groups.find_by(description=description)
        if len(groups) == 0:
            return None
        else:
            return groups[0]

    def tenant(self, tenant_id, tenant_name):
        """Search for tenant entity by name or id
        Returns:
            the tenant entity, None if no id or name was supplied
        """

        if tenant_id:
            tenant = self.client.get_entity("tenants", tenant_id)
            if not tenant:
                self.module.fail_json(msg=f"Tenant with id '{tenant_id}' not found in manageiq")
            return tenant
        else:
            if tenant_name:
                tenant_res = self.client.collections.tenants.find_by(name=tenant_name)
                if not tenant_res:
                    self.module.fail_json(msg=f"Tenant '{tenant_name}' not found in manageiq")
                if len(tenant_res) > 1:
                    self.module.fail_json(msg=f"Multiple tenants found in manageiq with name '{tenant_name}'")
                tenant = tenant_res[0]
                return tenant
            else:
                # No tenant name or tenant id supplied
                return None

    def role(self, role_id, role_name):
        """Search for a role object by name or id.
        Returns:
            the role entity, None no id or name was supplied

            the role, or send a module Fail signal if role not found.
        """
        if role_id:
            role = self.client.get_entity("roles", role_id)
            if not role:
                self.module.fail_json(msg=f"Role with id '{role_id}' not found in manageiq")
            return role
        else:
            if role_name:
                role_res = self.client.collections.roles.find_by(name=role_name)
                if not role_res:
                    self.module.fail_json(msg=f"Role '{role_name}' not found in manageiq")
                if len(role_res) > 1:
                    self.module.fail_json(msg=f"Multiple roles found in manageiq with name '{role_name}'")
                return role_res[0]
            else:
                # No role name or role id supplied
                return None

    @staticmethod
    def merge_dict_values(norm_current_values, norm_updated_values):
        """Create an merged update object for manageiq group filters.

            The input dict contain the tag values per category.
            If the new values contain the category, all tags for that category are replaced
            If the new values do not contain the category, the existing tags are kept

        Returns:
            the nested array with the merged values, used in the update post body
        """

        # If no updated values are supplied, in merge mode, the original values must be returned
        # otherwise the existing tag filters will be removed.
        if norm_current_values and (not norm_updated_values):
            return norm_current_values

        # If no existing tag filters exist, use the user supplied values
        if (not norm_current_values) and norm_updated_values:
            return norm_updated_values

        # start with norm_current_values's keys and values
        res = norm_current_values.copy()
        # replace res with norm_updated_values's keys and values
        res.update(norm_updated_values)
        return res

    def delete_group(self, group):
        """Deletes a group from manageiq.

        Returns:
            a dict of:
            changed: boolean indicating if the entity was updated.
            msg: a short message describing the operation executed.
        """
        try:
            url = f"{self.api_url}/groups/{group['id']}"
            result = self.client.post(url, action="delete")
        except Exception as e:
            self.module.fail_json(msg=f"failed to delete group {group['description']}: {e}")

        if result["success"] is False:
            self.module.fail_json(msg=result["message"])

        return dict(changed=True, msg=f"deleted group {group['description']} with id {group['id']}")

    def edit_group(
        self,
        group,
        description,
        role,
        tenant,
        norm_managed_filters,
        managed_filters_merge_mode,
        belongsto_filters,
        belongsto_filters_merge_mode,
    ):
        """Edit a manageiq group.

        Returns:
            a dict of:
            changed: boolean indicating if the entity was updated.
            msg: a short message describing the operation executed.
        """

        if role or norm_managed_filters or belongsto_filters:
            group.reload(attributes=["miq_user_role_name", "entitlement"])

        try:
            current_role = group["miq_user_role_name"]
        except AttributeError:
            current_role = None

        changed = False
        resource = {}

        if description and group["description"] != description:
            resource["description"] = description
            changed = True

        if tenant and group["tenant_id"] != tenant["id"]:
            resource["tenant"] = dict(id=tenant["id"])
            changed = True

        if role and current_role != role["name"]:
            resource["role"] = dict(id=role["id"])
            changed = True

        if norm_managed_filters or belongsto_filters:
            # Only compare if filters are supplied
            entitlement = group["entitlement"]

            if "filters" not in entitlement:
                # No existing filters exist, use supplied filters
                managed_tag_filters_post_body = self.normalized_managed_tag_filters_to_miq(norm_managed_filters)
                resource["filters"] = {"managed": managed_tag_filters_post_body, "belongsto": belongsto_filters}
                changed = True
            else:
                current_filters = entitlement["filters"]
                new_filters = self.edit_group_edit_filters(
                    current_filters,
                    norm_managed_filters,
                    managed_filters_merge_mode,
                    belongsto_filters,
                    belongsto_filters_merge_mode,
                )
                if new_filters:
                    resource["filters"] = new_filters
                    changed = True

        if not changed:
            return dict(changed=False, msg=f"group {group['description']} is not changed.")

        # try to update group
        try:
            self.client.post(group["href"], action="edit", resource=resource)
            changed = True
        except Exception as e:
            self.module.fail_json(msg=f"failed to update group {group['name']}: {e!s}")

        return dict(changed=changed, msg=f"successfully updated the group {group['description']} with id {group['id']}")

    def edit_group_edit_filters(
        self,
        current_filters,
        norm_managed_filters,
        managed_filters_merge_mode,
        belongsto_filters,
        belongsto_filters_merge_mode,
    ):
        """Edit a manageiq group filters.

        Returns:
            None if no the group was not updated
            If the group was updated the post body part for updating the group
        """
        filters_updated = False
        new_filters_resource = {}

        current_belongsto_set = current_filters.get("belongsto", set())

        if belongsto_filters:
            new_belongsto_set = set(belongsto_filters)
        else:
            new_belongsto_set = set()

        if current_belongsto_set == new_belongsto_set:
            new_filters_resource["belongsto"] = current_filters["belongsto"]
        else:
            if belongsto_filters_merge_mode == "merge":
                current_belongsto_set.update(new_belongsto_set)
                new_filters_resource["belongsto"] = list(current_belongsto_set)
            else:
                new_filters_resource["belongsto"] = list(new_belongsto_set)
            filters_updated = True

        # Process belongsto managed filter tags
        # The input is in the form dict with keys are the categories and the tags are supplied string array
        # ManageIQ, the current_managed, uses an array of arrays. One array of categories.
        # We normalize the user input from a dict with arrays to a dict of sorted arrays
        # We normalize the current manageiq array of arrays also to a dict of sorted arrays so we can compare
        norm_current_filters = self.manageiq_filters_to_sorted_dict(current_filters)

        if norm_current_filters == norm_managed_filters:
            if "managed" in current_filters:
                new_filters_resource["managed"] = current_filters["managed"]
        else:
            if managed_filters_merge_mode == "merge":
                merged_dict = self.merge_dict_values(norm_current_filters, norm_managed_filters)
                new_filters_resource["managed"] = self.normalized_managed_tag_filters_to_miq(merged_dict)
            else:
                new_filters_resource["managed"] = self.normalized_managed_tag_filters_to_miq(norm_managed_filters)
            filters_updated = True

        if not filters_updated:
            return None

        return new_filters_resource

    def create_group(self, description, role, tenant, norm_managed_filters, belongsto_filters):
        """Creates the group in manageiq.

        Returns:
            the created group id, name, created_on timestamp,
            updated_on timestamp.
        """
        # check for required arguments
        for key, value in dict(description=description).items():
            if value in (None, ""):
                self.module.fail_json(msg=f"missing required argument: {key}")

        url = f"{self.api_url}/groups"

        resource = {"description": description}

        if role is not None:
            resource["role"] = dict(id=role["id"])

        if tenant is not None:
            resource["tenant"] = dict(id=tenant["id"])

        if norm_managed_filters or belongsto_filters:
            managed_tag_filters_post_body = self.normalized_managed_tag_filters_to_miq(norm_managed_filters)
            resource["filters"] = {"managed": managed_tag_filters_post_body, "belongsto": belongsto_filters}

        try:
            result = self.client.post(url, action="create", resource=resource)
        except Exception as e:
            self.module.fail_json(msg=f"failed to create group {description}: {e}")

        return dict(changed=True, msg=f"successfully created group {description}", group_id=result["results"][0]["id"])

    @staticmethod
    def normalized_managed_tag_filters_to_miq(norm_managed_filters):
        if not norm_managed_filters:
            return None

        return list(norm_managed_filters.values())

    @staticmethod
    def manageiq_filters_to_sorted_dict(current_filters):
        current_managed_filters = current_filters.get("managed")
        if not current_managed_filters:
            return None

        res = {}
        for tag_list in current_managed_filters:
            tag_list.sort()
            key = tag_list[0].split("/")[2]
            res[key] = tag_list

        return res

    @staticmethod
    def normalize_user_managed_filters_to_sorted_dict(managed_filters, module):
        if not managed_filters:
            return None

        res = {}
        for cat_key in managed_filters:
            cat_array = []
            if not isinstance(managed_filters[cat_key], list):
                module.fail_json(msg=f'Entry "{cat_key}" of managed_filters must be a list!')
            for tags in managed_filters[cat_key]:
                miq_managed_tag = f"/managed/{cat_key}/{tags}"
                cat_array.append(miq_managed_tag)
            # Do not add empty categories. ManageIQ will remove all categories that are not supplied
            if cat_array:
                cat_array.sort()
                res[cat_key] = cat_array
        return res

    @staticmethod
    def create_result_group(group):
        """Creates the ansible result object from a manageiq group entity

        Returns:
            a dict with the group id, description, role, tenant, filters, group_type, created_on, updated_on
        """
        try:
            role_name = group["miq_user_role_name"]
        except AttributeError:
            role_name = None

        managed_filters = None
        belongsto_filters = None
        if "filters" in group["entitlement"]:
            filters = group["entitlement"]["filters"]
            belongsto_filters = filters.get("belongsto")
            group_managed_filters = filters.get("managed")
            if group_managed_filters:
                managed_filters = {}
                for tag_list in group_managed_filters:
                    key = tag_list[0].split("/")[2]
                    tags = []
                    for t in tag_list:
                        tags.append(t.split("/")[3])
                    managed_filters[key] = tags

        return dict(
            id=group["id"],
            description=group["description"],
            role=role_name,
            tenant=group["tenant"]["name"],
            managed_filters=managed_filters,
            belongsto_filters=belongsto_filters,
            group_type=group["group_type"],
            created_on=group["created_on"],
            updated_on=group["updated_on"],
        )


def main():
    argument_spec = dict(
        description=dict(required=True, type="str"),
        state=dict(choices=["absent", "present"], default="present"),
        role_id=dict(type="int"),
        role=dict(type="str"),
        tenant_id=dict(type="int"),
        tenant=dict(type="str"),
        managed_filters=dict(type="dict"),
        managed_filters_merge_mode=dict(choices=["merge", "replace"], default="replace"),
        belongsto_filters=dict(type="list", elements="str"),
        belongsto_filters_merge_mode=dict(choices=["merge", "replace"], default="replace"),
    )
    # add the manageiq connection arguments to the arguments
    argument_spec.update(manageiq_argument_spec())

    module = AnsibleModule(argument_spec=argument_spec)

    description = module.params["description"]
    state = module.params["state"]
    role_id = module.params["role_id"]
    role_name = module.params["role"]
    tenant_id = module.params["tenant_id"]
    tenant_name = module.params["tenant"]
    managed_filters = module.params["managed_filters"]
    managed_filters_merge_mode = module.params["managed_filters_merge_mode"]
    belongsto_filters = module.params["belongsto_filters"]
    belongsto_filters_merge_mode = module.params["belongsto_filters_merge_mode"]

    manageiq = ManageIQ(module)
    manageiq_group = ManageIQgroup(manageiq)

    group = manageiq_group.group(description)

    # group should not exist
    if state == "absent":
        # if we have a group, delete it
        if group:
            res_args = manageiq_group.delete_group(group)
        # if we do not have a group, nothing to do
        else:
            res_args = dict(changed=False, msg=f"group '{description}' does not exist in manageiq")

    # group should exist
    if state == "present":
        tenant = manageiq_group.tenant(tenant_id, tenant_name)
        role = manageiq_group.role(role_id, role_name)
        norm_managed_filters = manageiq_group.normalize_user_managed_filters_to_sorted_dict(managed_filters, module)
        # if we have a group, edit it
        if group:
            res_args = manageiq_group.edit_group(
                group,
                description,
                role,
                tenant,
                norm_managed_filters,
                managed_filters_merge_mode,
                belongsto_filters,
                belongsto_filters_merge_mode,
            )

        # if we do not have a group, create it
        else:
            res_args = manageiq_group.create_group(description, role, tenant, norm_managed_filters, belongsto_filters)
            group = manageiq.client.get_entity("groups", res_args["group_id"])

        group.reload(expand="resources", attributes=["miq_user_role_name", "tenant", "entitlement"])
        res_args["group"] = manageiq_group.create_result_group(group)

    module.exit_json(**res_args)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2021, Sergey Mikhaltsov <metanovii@gmail.com>
# Copyright (c) 2020, Zainab Alsaffar <Zainab.Alsaffar@mail.rit.edu>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_project_members
short_description: Manage project members on GitLab Server
version_added: 2.2.0
description:
  - This module allows to add and remove members to/from a project, or change a member's access level in a project on GitLab.
author:
  - Sergey Mikhaltsov (@metanovii)
  - Zainab Alsaffar (@zanssa)
requirements:
  - python-gitlab python module <= 1.15.0
  - owner or maintainer rights to project on the GitLab server
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  project:
    description:
      - The name (or full path) of the GitLab project the member is added to/removed from.
    required: true
    type: str
  gitlab_user:
    description:
      - A username or a list of usernames to add to/remove from the GitLab project.
      - Mutually exclusive with O(gitlab_users_access).
    type: list
    elements: str
  access_level:
    description:
      - The access level for the user.
      - Required if O(state=present), user state is set to present.
      - V(owner) was added in community.general 10.6.0.
    type: str
    choices: ['guest', 'reporter', 'developer', 'maintainer', 'owner']
  gitlab_users_access:
    description:
      - Provide a list of user to access level mappings.
      - Every dictionary in this list specifies a user (by username) and the access level the user should have.
      - Mutually exclusive with O(gitlab_user) and O(access_level).
      - Use together with O(purge_users) to remove all users not specified here from the project.
    type: list
    elements: dict
    suboptions:
      name:
        description: A username or a list of usernames to add to/remove from the GitLab project.
        type: str
        required: true
      access_level:
        description:
          - The access level for the user.
          - Required if O(state=present), user state is set to present.
          - V(owner) was added in community.general 10.6.0.
        type: str
        choices: ['guest', 'reporter', 'developer', 'maintainer', 'owner']
        required: true
    version_added: 3.7.0
  state:
    description:
      - State of the member in the project.
      - On V(present), it adds a user to a GitLab project.
      - On V(absent), it removes a user from a GitLab project.
    choices: ['present', 'absent']
    default: 'present'
    type: str
  purge_users:
    description:
      - Adds/remove users of the given access_level to match the given O(gitlab_user)/O(gitlab_users_access) list. If omitted
        do not purge orphaned members.
      - Is only used when O(state=present).
      - V(owner) was added in community.general 10.6.0.
    type: list
    elements: str
    choices: ['guest', 'reporter', 'developer', 'maintainer', 'owner']
    version_added: 3.7.0
"""

EXAMPLES = r"""
- name: Add a user to a GitLab Project
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    validate_certs: true
    project: projectname
    gitlab_user: username
    access_level: developer
    state: present

- name: Remove a user from a GitLab project
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    project: projectname
    gitlab_user: username
    state: absent

- name: Add a list of Users to A GitLab project
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    gitlab_project: projectname
    gitlab_user:
      - user1
      - user2
    access_level: developer
    state: present

- name: Add a list of Users with Dedicated Access Levels to A GitLab project
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    project: projectname
    gitlab_users_access:
      - name: user1
        access_level: developer
      - name: user2
        access_level: maintainer
    state: present

- name: Add a user, remove all others which might be on this access level
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    project: projectname
    gitlab_user: username
    access_level: developer
    purge_users: developer
    state: present

- name: Remove a list of Users with Dedicated Access Levels to A GitLab project
  community.general.gitlab_project_members:
    api_url: 'https://gitlab.example.com'
    api_token: 'Your-Private-Token'
    project: projectname
    gitlab_users_access:
      - name: user1
        access_level: developer
      - name: user2
        access_level: maintainer
    state: absent
"""

RETURN = r""" # """

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    gitlab,
    gitlab_authentication,
)


class GitLabProjectMembers:
    def __init__(self, module, gl):
        self._module = module
        self._gitlab = gl

    def get_project(self, project_name):
        try:
            project_exists = self._gitlab.projects.get(project_name)
            return project_exists.id
        except gitlab.exceptions.GitlabGetError:
            project_exists = self._gitlab.projects.list(search=project_name, all=False)
            if len(project_exists) == 1:
                return project_exists[0].id
            if len(project_exists) > 1:
                self._module.fail_json(
                    msg=(
                        f"More than one project matches '{project_name}'. "
                        "Use the full path ('group/project') to disambiguate."
                    )
                )

    def get_user_id(self, gitlab_user):
        user_exists = self._gitlab.users.list(username=gitlab_user, all=False)
        if user_exists:
            return user_exists[0].id

    # get all members in a project
    def get_members_in_a_project(self, gitlab_project_id):
        project = self._gitlab.projects.get(gitlab_project_id)
        return project.members.list(all=True)

    # get single member in a project by user name
    def get_member_in_a_project(self, gitlab_project_id, gitlab_user_id):
        member = None
        project = self._gitlab.projects.get(gitlab_project_id)
        try:
            member = project.members.get(gitlab_user_id)
            if member:
                return member
        except gitlab.exceptions.GitlabGetError:
            return None

    # check if the user is a member of the project
    def is_user_a_member(self, members, gitlab_user_id):
        return any(member.id == gitlab_user_id for member in members)

    # add user to a project
    def add_member_to_project(self, gitlab_user_id, gitlab_project_id, access_level):
        project = self._gitlab.projects.get(gitlab_project_id)
        project.members.create({"user_id": gitlab_user_id, "access_level": access_level})

    # remove user from a project
    def remove_user_from_project(self, gitlab_user_id, gitlab_project_id):
        project = self._gitlab.projects.get(gitlab_project_id)
        project.members.delete(gitlab_user_id)

    # get user's access level
    def get_user_access_level(self, members, gitlab_user_id):
        for member in members:
            if member.id == gitlab_user_id:
                return member.access_level

    # update user's access level in a project
    def update_user_access_level(self, members, gitlab_user_id, access_level):
        for member in members:
            if member.id == gitlab_user_id:
                member.access_level = access_level
                member.save()


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            project=dict(type="str", required=True),
            gitlab_user=dict(type="list", elements="str"),
            state=dict(type="str", default="present", choices=["present", "absent"]),
            access_level=dict(type="str", choices=["guest", "reporter", "developer", "maintainer", "owner"]),
            purge_users=dict(
                type="list", elements="str", choices=["guest", "reporter", "developer", "maintainer", "owner"]
            ),
            gitlab_users_access=dict(
                type="list",
                elements="dict",
                options=dict(
                    name=dict(type="str", required=True),
                    access_level=dict(
                        type="str", choices=["guest", "reporter", "developer", "maintainer", "owner"], required=True
                    ),
                ),
            ),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
            ["gitlab_user", "gitlab_users_access"],
            ["access_level", "gitlab_users_access"],
        ],
        required_together=[
            ["api_username", "api_password"],
            ["gitlab_user", "access_level"],
        ],
        required_one_of=[
            ["api_username", "api_token", "api_oauth_token", "api_job_token"],
            ["gitlab_user", "gitlab_users_access"],
        ],
        required_if=[
            ["state", "present", ["access_level", "gitlab_users_access"], True],
        ],
        supports_check_mode=True,
    )

    # check prerequisites and connect to gitlab server
    gl = gitlab_authentication(module)

    access_level_int = {
        "guest": gitlab.const.GUEST_ACCESS,
        "reporter": gitlab.const.REPORTER_ACCESS,
        "developer": gitlab.const.DEVELOPER_ACCESS,
        "maintainer": gitlab.const.MAINTAINER_ACCESS,
        "owner": gitlab.const.OWNER_ACCESS,
    }

    gitlab_project = module.params["project"]
    state = module.params["state"]
    access_level = module.params["access_level"]
    purge_users = module.params["purge_users"]

    if purge_users:
        purge_users = [access_level_int[level] for level in purge_users]

    project = GitLabProjectMembers(module, gl)

    gitlab_project_id = project.get_project(gitlab_project)

    # project doesn't exist
    if not gitlab_project_id:
        module.fail_json(msg=f"project '{gitlab_project}' not found.")

    members = []
    if module.params["gitlab_user"] is not None:
        gitlab_users_access = []
        gitlab_users = module.params["gitlab_user"]
        for gl_user in gitlab_users:
            gitlab_users_access.append(
                {"name": gl_user, "access_level": access_level_int[access_level] if access_level else None}
            )
    elif module.params["gitlab_users_access"] is not None:
        gitlab_users_access = module.params["gitlab_users_access"]
        for user_level in gitlab_users_access:
            user_level["access_level"] = access_level_int[user_level["access_level"]]

    if len(gitlab_users_access) == 1 and not purge_users:
        # only single user given
        members = [
            project.get_member_in_a_project(gitlab_project_id, project.get_user_id(gitlab_users_access[0]["name"]))
        ]
        if members[0] is None:
            members = []
    elif len(gitlab_users_access) > 1 or purge_users:
        # list of users given
        members = project.get_members_in_a_project(gitlab_project_id)
    else:
        module.exit_json(
            changed="OK", result="Nothing to do, please give at least one user or set purge_users true.", result_data=[]
        )

    changed = False
    error = False
    changed_users = []
    changed_data = []

    for gitlab_user in gitlab_users_access:
        gitlab_user_id = project.get_user_id(gitlab_user["name"])

        # user doesn't exist
        if not gitlab_user_id:
            if state == "absent":
                changed_users.append(f"user '{gitlab_user['name']}' not found, and thus also not part of the project")
                changed_data.append(
                    {
                        "gitlab_user": gitlab_user["name"],
                        "result": "OK",
                        "msg": f"user '{gitlab_user['name']}' not found, and thus also not part of the project",
                    }
                )
            else:
                error = True
                changed_users.append(f"user '{gitlab_user['name']}' not found.")
                changed_data.append(
                    {
                        "gitlab_user": gitlab_user["name"],
                        "result": "FAILED",
                        "msg": f"user '{gitlab_user['name']}' not found.",
                    }
                )
            continue

        is_user_a_member = project.is_user_a_member(members, gitlab_user_id)

        # check if the user is a member in the project
        if not is_user_a_member:
            if state == "present":
                # add user to the project
                try:
                    if not module.check_mode:
                        project.add_member_to_project(gitlab_user_id, gitlab_project_id, gitlab_user["access_level"])
                    changed = True
                    changed_users.append(f"Successfully added user '{gitlab_user['name']}' to project")
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "CHANGED",
                            "msg": f"Successfully added user '{gitlab_user['name']}' to project",
                        }
                    )
                except gitlab.exceptions.GitlabCreateError as e:
                    error = True
                    changed_users.append(f"Failed to updated the access level for the user, '{gitlab_user['name']}'")
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "FAILED",
                            "msg": f"Not allowed to add the access level for the member, {gitlab_user['name']}: {e}",
                        }
                    )
            # state as absent
            else:
                changed_users.append(
                    f"User, '{gitlab_user['name']}', is not a member in the project. No change to report"
                )
                changed_data.append(
                    {
                        "gitlab_user": gitlab_user["name"],
                        "result": "OK",
                        "msg": f"User, '{gitlab_user['name']}', is not a member in the project. No change to report",
                    }
                )
        # in case that a user is a member
        else:
            if state == "present":
                # compare the access level
                user_access_level = project.get_user_access_level(members, gitlab_user_id)
                if user_access_level == gitlab_user["access_level"]:
                    changed_users.append(
                        f"User, '{gitlab_user['name']}', is already a member in the project. No change to report"
                    )
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "OK",
                            "msg": f"User, '{gitlab_user['name']}', is already a member in the project. No change to report",
                        }
                    )
                else:
                    # update the access level for the user
                    try:
                        if not module.check_mode:
                            project.update_user_access_level(members, gitlab_user_id, gitlab_user["access_level"])
                        changed = True
                        changed_users.append(
                            f"Successfully updated the access level for the user, '{gitlab_user['name']}'"
                        )
                        changed_data.append(
                            {
                                "gitlab_user": gitlab_user["name"],
                                "result": "CHANGED",
                                "msg": f"Successfully updated the access level for the user, '{gitlab_user['name']}'",
                            }
                        )
                    except gitlab.exceptions.GitlabUpdateError as e:
                        error = True
                        changed_users.append(
                            f"Failed to updated the access level for the user, '{gitlab_user['name']}'"
                        )
                        changed_data.append(
                            {
                                "gitlab_user": gitlab_user["name"],
                                "result": "FAILED",
                                "msg": f"Not allowed to update the access level for the member, {gitlab_user['name']}: {e}",
                            }
                        )
            else:
                # remove the user from the project
                try:
                    if not module.check_mode:
                        project.remove_user_from_project(gitlab_user_id, gitlab_project_id)
                    changed = True
                    changed_users.append(f"Successfully removed user, '{gitlab_user['name']}', from the project")
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "CHANGED",
                            "msg": f"Successfully removed user, '{gitlab_user['name']}', from the project",
                        }
                    )
                except gitlab.exceptions.GitlabDeleteError as e:
                    error = True
                    changed_users.append(f"Failed to removed user, '{gitlab_user['name']}', from the project")
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "FAILED",
                            "msg": f"Failed to remove user, '{gitlab_user['name']}' from the project: {e}",
                        }
                    )

    # if state = present and purge_users set delete users which are in members having give access level but not in gitlab_users
    if state == "present" and purge_users:
        uppercase_names_in_gitlab_users_access = []
        for name in gitlab_users_access:
            uppercase_names_in_gitlab_users_access.append(name["name"].upper())

        for member in members:
            if (
                member.access_level in purge_users
                and member.username.upper() not in uppercase_names_in_gitlab_users_access
            ):
                try:
                    if not module.check_mode:
                        project.remove_user_from_project(member.id, gitlab_project_id)
                    changed = True
                    changed_users.append(
                        f"Successfully removed user '{member.username}', from project. Was not in given list"
                    )
                    changed_data.append(
                        {
                            "gitlab_user": member.username,
                            "result": "CHANGED",
                            "msg": f"Successfully removed user '{member.username}', from project. Was not in given list",
                        }
                    )
                except gitlab.exceptions.GitlabDeleteError as e:
                    error = True
                    changed_users.append(f"Failed to removed user, '{gitlab_user['name']}', from the project")
                    changed_data.append(
                        {
                            "gitlab_user": gitlab_user["name"],
                            "result": "FAILED",
                            "msg": f"Failed to remove user, '{gitlab_user['name']}' from the project: {e}",
                        }
                    )

    if len(gitlab_users_access) == 1 and error:
        # if single user given and an error occurred return error for list errors will be per user
        module.fail_json(msg=f"FAILED: '{changed_users[0]} '", result_data=changed_data)
    elif error:
        module.fail_json(msg="FAILED: At least one given user/permission could not be set", result_data=changed_data)

    module.exit_json(
        changed=changed, msg="Successfully set memberships", result="\n".join(changed_users), result_data=changed_data
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: crypto_info
author: "Felix Fontein (@felixfontein)"
short_description: Retrieve cryptographic capabilities
version_added: 2.1.0
description:
  - Retrieve information on cryptographic capabilities.
  - The current version retrieves information on the L(Python cryptography library, https://cryptography.io/) available to
    Ansible modules, and on the OpenSSL binary C(openssl) found in the path.
extends_documentation_fragment:
  - community.crypto._attributes
  - community.crypto._attributes.info_module
  - community.crypto._attributes.idempotent_not_modify_state
options: {}
"""

EXAMPLES = r"""
---
- name: Retrieve information
  community.crypto.crypto_info:
    account_key_src: /etc/pki/cert/private/account.key
  register: crypto_information

- name: Show retrieved information
  ansible.builtin.debug:
    var: crypto_information
"""

RETURN = r"""
python_cryptography_installed:
  description: Whether the L(Python cryptography library, https://cryptography.io/) is installed.
  returned: always
  type: bool
  sample: true

python_cryptography_import_error:
  description: Import error when trying to import the L(Python cryptography library, https://cryptography.io/).
  returned: when RV(python_cryptography_installed=false)
  type: str

python_cryptography_capabilities:
  description: Information on the installed L(Python cryptography library, https://cryptography.io/).
  returned: when RV(python_cryptography_installed=true)
  type: dict
  contains:
    version:
      description: The library version.
      type: str
    curves:
      description:
        - List of all supported elliptic curves.
        - Theoretically this should be non-empty for version 0.5 and higher, depending on the libssl version used.
      type: list
      elements: str
    has_ec:
      description:
        - Whether elliptic curves are supported.
        - Theoretically this should be the case for version 0.5 and higher, depending on the libssl version used.
      type: bool
    has_ec_sign:
      description:
        - Whether signing with elliptic curves is supported.
        - Theoretically this should be the case for version 1.5 and higher, depending on the libssl version used.
      type: bool
    has_ed25519:
      description:
        - Whether Ed25519 keys are supported.
        - Theoretically this should be the case for version 2.6 and higher, depending on the libssl version used.
      type: bool
    has_ed25519_sign:
      description:
        - Whether signing with Ed25519 keys is supported.
        - Theoretically this should be the case for version 2.6 and higher, depending on the libssl version used.
      type: bool
    has_ed448:
      description:
        - Whether Ed448 keys are supported.
        - Theoretically this should be the case for version 2.6 and higher, depending on the libssl version used.
      type: bool
    has_ed448_sign:
      description:
        - Whether signing with Ed448 keys is supported.
        - Theoretically this should be the case for version 2.6 and higher, depending on the libssl version used.
      type: bool
    has_dsa:
      description:
        - Whether DSA keys are supported.
        - Theoretically this should be the case for version 0.5 and higher.
      type: bool
    has_dsa_sign:
      description:
        - Whether signing with DSA keys is supported.
        - Theoretically this should be the case for version 1.5 and higher.
      type: bool
    has_rsa:
      description:
        - Whether RSA keys are supported.
        - Theoretically this should be the case for version 0.5 and higher.
      type: bool
    has_rsa_sign:
      description:
        - Whether signing with RSA keys is supported.
        - Theoretically this should be the case for version 1.4 and higher.
      type: bool
    has_x25519:
      description:
        - Whether X25519 keys are supported.
        - Theoretically this should be the case for version 2.0 and higher, depending on the libssl version used.
      type: bool
    has_x25519_serialization:
      description:
        - Whether serialization of X25519 keys is supported.
        - Theoretically this should be the case for version 2.5 and higher, depending on the libssl version used.
      type: bool
    has_x448:
      description:
        - Whether X448 keys are supported.
        - Theoretically this should be the case for version 2.5 and higher, depending on the libssl version used.
      type: bool

openssl_present:
  description: Whether the OpenSSL binary C(openssl) is installed and can be found in the PATH.
  returned: always
  type: bool
  sample: true

openssl:
  description: Information on the installed OpenSSL binary.
  returned: when RV(openssl_present=true)
  type: dict
  contains:
    path:
      description: Path of the OpenSSL binary.
      type: str
      sample: /usr/bin/openssl
    version:
      description: The OpenSSL version.
      type: str
      sample: 1.1.1m
    version_output:
      description: The complete output of C(openssl version).
      type: str
      sample: 'OpenSSL 1.1.1m  14 Dec 2021\n'
"""

import traceback
import typing as t

from ansible.module_utils.basic import AnsibleModule

CRYPTOGRAPHY_VERSION: str | None
CRYPTOGRAPHY_IMP_ERR: str | None
try:
    import cryptography
    import cryptography.hazmat.primitives.asymmetric
    from cryptography.exceptions import UnsupportedAlgorithm

    try:
        # While UnsupportedAlgorithm got added in cryptography 0.1, InternalError
        # only got added in 0.2, so let's guard the import
        from cryptography.exceptions import InternalError as CryptographyInternalError
    except ImportError:
        CryptographyInternalError = Exception  # type: ignore
except ImportError:
    UnsupportedAlgorithm = Exception  # type: ignore
    CryptographyInternalError = Exception  # type: ignore
    HAS_CRYPTOGRAPHY = False
    CRYPTOGRAPHY_VERSION = None  # pylint: disable=invalid-name
    CRYPTOGRAPHY_IMP_ERR = traceback.format_exc()  # pylint: disable=invalid-name
else:
    HAS_CRYPTOGRAPHY = True
    CRYPTOGRAPHY_VERSION = cryptography.__version__  # pylint: disable=invalid-name
    CRYPTOGRAPHY_IMP_ERR = None  # pylint: disable=invalid-name


CURVES = (
    ("secp224r1", "SECP224R1"),
    ("secp256k1", "SECP256K1"),
    ("secp256r1", "SECP256R1"),
    ("secp384r1", "SECP384R1"),
    ("secp521r1", "SECP521R1"),
    ("secp192r1", "SECP192R1"),
    ("sect163k1", "SECT163K1"),
    ("sect163r2", "SECT163R2"),
    ("sect233k1", "SECT233K1"),
    ("sect233r1", "SECT233R1"),
    ("sect283k1", "SECT283K1"),
    ("sect283r1", "SECT283R1"),
    ("sect409k1", "SECT409K1"),
    ("sect409r1", "SECT409R1"),
    ("sect571k1", "SECT571K1"),
    ("sect571r1", "SECT571R1"),
    ("brainpoolP256r1", "BrainpoolP256R1"),
    ("brainpoolP384r1", "BrainpoolP384R1"),
    ("brainpoolP512r1", "BrainpoolP512R1"),
)


def add_crypto_information(module: AnsibleModule) -> dict[str, t.Any]:
    result: dict[str, t.Any] = {}
    result["python_cryptography_installed"] = HAS_CRYPTOGRAPHY
    if not HAS_CRYPTOGRAPHY:
        result["python_cryptography_import_error"] = CRYPTOGRAPHY_IMP_ERR
        return result

    # Test for DSA
    has_dsa = False
    has_dsa_sign = False
    try:
        # added in 0.5 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/dsa/
        from cryptography.hazmat.primitives.asymmetric import dsa

        has_dsa = True
        try:
            # added later in 1.5
            dsa.DSAPrivateKey.sign  # noqa: B018 # pylint: disable=pointless-statement
            has_dsa_sign = True
        except AttributeError:
            pass
    except ImportError:
        pass

    # Test for RSA
    has_rsa = False
    has_rsa_sign = False
    try:
        # added in 0.5 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/rsa/
        from cryptography.hazmat.primitives.asymmetric import rsa

        has_rsa = True
        try:
            # added later in 1.4
            rsa.RSAPrivateKey.sign  # noqa: B018 # pylint: disable=pointless-statement
            has_rsa_sign = True
        except AttributeError:
            pass
    except ImportError:
        pass

    # Test for Ed25519
    has_ed25519 = False
    has_ed25519_sign = False
    try:
        # added in 2.6 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/ed25519/
        from cryptography.hazmat.primitives.asymmetric import ed25519

        try:
            ed25519.Ed25519PrivateKey.from_private_bytes(b"")
        except ValueError:
            pass

        has_ed25519 = True
        try:
            # added with the primitive in 2.6
            ed25519.Ed25519PrivateKey.sign  # noqa: B018 # pylint: disable=pointless-statement
            has_ed25519_sign = True
        except AttributeError:
            pass
    except (ImportError, UnsupportedAlgorithm):
        pass

    # Test for Ed448
    has_ed448 = False
    has_ed448_sign = False
    try:
        # added in 2.6 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/ed448/
        from cryptography.hazmat.primitives.asymmetric import ed448

        try:
            ed448.Ed448PrivateKey.from_private_bytes(b"")
        except ValueError:
            pass

        has_ed448 = True
        try:
            # added with the primitive in 2.6
            ed448.Ed448PrivateKey.sign  # noqa: B018 # pylint: disable=pointless-statement
            has_ed448_sign = True
        except AttributeError:
            pass
    except (ImportError, UnsupportedAlgorithm):
        pass

    # Test for X25519
    has_x25519 = False
    has_x25519_full = False
    try:
        # added in 2.0 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/x25519/
        from cryptography.hazmat.primitives.asymmetric import x25519

        try:
            # added later in 2.5
            x25519.X25519PrivateKey.private_bytes  # noqa: B018 # pylint: disable=pointless-statement
            full = True
        except AttributeError:
            full = False

        try:
            if full:
                x25519.X25519PrivateKey.from_private_bytes(b"")
            else:
                # Some versions do not support serialization and deserialization - use generate() instead
                x25519.X25519PrivateKey.generate()
        except ValueError:
            pass

        has_x25519 = True
        has_x25519_full = full
    except (ImportError, UnsupportedAlgorithm):
        pass

    # Test for X448
    has_x448 = False
    try:
        # added in 2.5 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/x448/
        from cryptography.hazmat.primitives.asymmetric import x448

        try:
            x448.X448PrivateKey.from_private_bytes(b"")
        except ValueError:
            pass

        has_x448 = True
    except (ImportError, UnsupportedAlgorithm):
        pass

    # Test for ECC
    has_ec = False
    has_ec_sign = False
    curves = []
    try:
        # added in 0.5 - https://cryptography.io/en/latest/hazmat/primitives/asymmetric/ec/
        from cryptography.hazmat.primitives.asymmetric import ec

        has_ec = True
        try:
            # added later in 1.5
            ec.EllipticCurvePrivateKey.sign  # noqa: B018 # pylint: disable=pointless-statement
            has_ec_sign = True
        except AttributeError:
            pass
    except ImportError:
        pass
    else:
        for curve_name, constructor_name in CURVES:
            ecclass = getattr(ec, constructor_name, None)
            if ecclass is not None:
                try:
                    ec.generate_private_key(curve=ecclass())
                    curves.append(curve_name)
                except UnsupportedAlgorithm:
                    pass
                except (  # pylint: disable=duplicate-except,bad-except-order
                    CryptographyInternalError
                ):
                    # On Fedora 41, some curves result in InternalError. This is probably because
                    # Fedora's cryptography is linked against the system libssl, which has the
                    # curves removed.
                    pass

    # Compose result
    info = {
        "version": CRYPTOGRAPHY_VERSION,
        "curves": curves,
        "has_ec": has_ec,
        "has_ec_sign": has_ec_sign,
        "has_ed25519": has_ed25519,
        "has_ed25519_sign": has_ed25519_sign,
        "has_ed448": has_ed448,
        "has_ed448_sign": has_ed448_sign,
        "has_dsa": has_dsa,
        "has_dsa_sign": has_dsa_sign,
        "has_rsa": has_rsa,
        "has_rsa_sign": has_rsa_sign,
        "has_x25519": has_x25519,
        "has_x25519_serialization": has_x25519 and has_x25519_full,
        "has_x448": has_x448,
    }
    result["python_cryptography_capabilities"] = info
    return result


def add_openssl_information(module: AnsibleModule) -> dict[str, t.Any]:
    openssl_binary = module.get_bin_path("openssl")
    result: dict[str, t.Any] = {
        "openssl_present": openssl_binary is not None,
    }
    if openssl_binary is None:
        return result

    openssl_result = {
        "path": openssl_binary,
    }
    result["openssl"] = openssl_result

    rc, out, _err = module.run_command([openssl_binary, "version"])
    if rc == 0:
        openssl_result["version_output"] = out
        parts = out.split(None, 2)
        if len(parts) > 1:
            openssl_result["version"] = parts[1]

    return result


INFO_FUNCTIONS = (
    add_crypto_information,
    add_openssl_information,
)


def main() -> t.NoReturn:
    module = AnsibleModule(argument_spec={}, supports_check_mode=True)
    result: dict[str, t.Any] = {}
    for fn in INFO_FUNCTIONS:
        result.update(fn(module))
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_adaptive_policy_policies
short_description: Resource module for organizations _adaptive _policy _policies
description:
  - Manage operations create, update and delete of the resource organizations _adaptive
    _policy _policies.
  - Add an Adaptive Policy.
  - Delete an Adaptive Policy.
  - Update an Adaptive Policy.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  acls:
    description: An ordered array of adaptive policy ACLs (each requires one unique
      attribute) that apply to this policy (default ).
    elements: dict
    suboptions:
      id:
        description: The ID of the adaptive policy ACL.
        type: str
      name:
        description: The name of the adaptive policy ACL.
        type: str
    type: list
  destinationGroup:
    description: The destination adaptive policy group (requires one unique attribute).
    suboptions:
      id:
        description: The ID of the destination adaptive policy group.
        type: str
      name:
        description: The name of the destination adaptive policy group.
        type: str
      sgt:
        description: The SGT of the destination adaptive policy group.
        type: int
    type: dict
  id:
    description: Id path parameter.
    type: str
  lastEntryRule:
    description: The rule to apply if there is no matching ACL (default "default").
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  sourceGroup:
    description: The source adaptive policy group (requires one unique attribute).
    suboptions:
      id:
        description: The ID of the source adaptive policy group.
        type: str
      name:
        description: The name of the source adaptive policy group.
        type: str
      sgt:
        description: The SGT of the source adaptive policy group.
        type: int
    type: dict
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationAdaptivePolicyPolicy
    description: Complete reference of the createOrganizationAdaptivePolicyPolicy
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-adaptive-policy-policy
  - name: Cisco Meraki documentation for organizations deleteOrganizationAdaptivePolicyPolicy
    description: Complete reference of the deleteOrganizationAdaptivePolicyPolicy
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-adaptive-policy-policy
  - name: Cisco Meraki documentation for organizations updateOrganizationAdaptivePolicyPolicy
    description: Complete reference of the updateOrganizationAdaptivePolicyPolicy
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-adaptive-policy-policy
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_adaptive_policy_policy,
    organizations.Organizations.delete_organization_adaptive_policy_policy,
    organizations.Organizations.update_organization_adaptive_policy_policy,
  - Paths used are
    post /organizations/{organizationId}/adaptivePolicy/policies,
    delete
    /organizations/{organizationId}/adaptivePolicy/policies/{id},
    put /organizations/{organizationId}/adaptivePolicy/policies/{id},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_adaptive_policy_policies:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    acls:
      - id: '444'
        name: Block web
    destinationGroup:
      id: '333'
      name: IoT Servers
      sgt: 51
    lastEntryRule: allow
    organizationId: string
    sourceGroup:
      id: '222'
      name: IoT Devices
      sgt: 50
- name: Update by id
  cisco.meraki.organizations_adaptive_policy_policies:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    acls:
      - id: '444'
        name: Block web
    destinationGroup:
      id: '333'
      name: IoT Servers
      sgt: 51
    id: string
    lastEntryRule: allow
    organizationId: string
    sourceGroup:
      id: '222'
      name: IoT Devices
      sgt: 50
- name: Delete by id
  cisco.meraki.organizations_adaptive_policy_policies:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    id: string
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "adaptivePolicyId": "string",
      "sourceGroup": {
        "id": "string",
        "name": "string",
        "sgt": 0
      },
      "destinationGroup": {
        "id": "string",
        "name": "string",
        "sgt": 0
      },
      "acls": [
        {
          "id": "string",
          "name": "string"
        }
      ],
      "lastEntryRule": "string",
      "createdAt": "string",
      "updatedAt": "string"
    }
"""



# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""The community.general.random_words Ansible lookup plugin."""

from __future__ import annotations

DOCUMENTATION = r"""
name: random_words
author:
  - Thomas Sjögren (@konstruktoid)
short_description: Return a number of random words
version_added: "4.0.0"
requirements:
  - xkcdpass U(https://github.com/redacted/XKCD-password-generator)
description:
  - Returns a number of random words. The output can for example be used for passwords.
  - See U(https://xkcd.com/936/) for background.
options:
  numwords:
    description:
      - The number of words.
    default: 6
    type: int
  min_length:
    description:
      - Minimum length of words to make password.
    default: 5
    type: int
  max_length:
    description:
      - Maximum length of words to make password.
    default: 9
    type: int
  delimiter:
    description:
      - The delimiter character between words.
    default: " "
    type: str
  case:
    description:
      - The method for setting the case of each word in the passphrase.
    choices: ["alternating", "upper", "lower", "random", "capitalize"]
    default: "lower"
    type: str
"""

EXAMPLES = r"""
- name: Generate password with default settings
  ansible.builtin.debug:
    var: lookup('community.general.random_words')
  # Example result: 'traitor gigabyte cesarean unless aspect clear'

- name: Generate password with six, five character, words
  ansible.builtin.debug:
    var: lookup('community.general.random_words', min_length=5, max_length=5)
  # Example result: 'brink banjo getup staff trump comfy'

- name: Generate password with three capitalized words and the '-' delimiter
  ansible.builtin.debug:
    var: lookup('community.general.random_words', numwords=3, delimiter='-', case='capitalize')
  # Example result: 'Overlabor-Faucet-Coastline'

- name: Generate password with three words without any delimiter
  ansible.builtin.debug:
    var: lookup('community.general.random_words', numwords=3, delimiter='')
  # Example result: 'deskworkmonopolystriking'
  # https://www.ncsc.gov.uk/blog-post/the-logic-behind-three-random-words
"""

RETURN = r"""
_raw:
  description: A single-element list containing random words.
  type: list
  elements: str
"""

from ansible.errors import AnsibleLookupError
from ansible.plugins.lookup import LookupBase

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_no_terms

try:
    from xkcdpass import xkcd_password as xp

    HAS_XKCDPASS = True
except ImportError:
    HAS_XKCDPASS = False


class LookupModule(LookupBase):
    """The random_words Ansible lookup class."""

    def run(self, terms, variables=None, **kwargs):
        if not HAS_XKCDPASS:
            raise AnsibleLookupError('Python xkcdpass library is required. Please install using "pip install xkcdpass"')

        self.set_options(var_options=variables, direct=kwargs)
        check_for_no_terms(self, terms=terms, direct=kwargs)
        method = self.get_option("case")
        delimiter = self.get_option("delimiter")
        max_length = self.get_option("max_length")
        min_length = self.get_option("min_length")
        numwords = self.get_option("numwords")

        words = xp.locate_wordfile()
        wordlist = xp.generate_wordlist(max_length=max_length, min_length=min_length, wordfile=words)

        values = xp.generate_xkcdpassword(wordlist, case=method, delimiter=delimiter, numwords=numwords)

        return [values]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: ec2_import_image
version_added: 7.0.0
short_description: Manage AWS EC2 import image tasks
description:
  - Import single or multi-volume disk images or EBS snapshots into an Amazon Machine Image (AMI).
  - Cancel an in-process import virtual machine task.
options:
  state:
    description:
      - Use O(state=present) to import single or multi-volume disk images or EBS snapshots into an Amazon Machine Image (AMI).
      - Use O(state=absent) to cancel an in-process import virtual machine task.
    default: "present"
    choices: ["present", "absent"]
    type: str
  task_name:
    description:
      - The name of the EC2 image import task.
    type: str
    aliases: ["name"]
    required: true
  architecture:
    description:
      - The architecture of the virtual machine.
    type: str
    choices: ["i386", "x86_64"]
  client_data:
    description:
      - The client-specific data.
    type: dict
    suboptions:
      comment:
        description:
          - A user-defined comment about the disk upload.
        type: str
      upload_end:
        description:
          - The time that the disk upload ends.
        type: str
      upload_size:
        description:
          - The size of the uploaded disk image, in GiB.
        type: float
      upload_start:
        description:
          - The time that the disk upload starts.
        type: str
  description:
    description:
      - A description string for the import image task.
    type: str
  disk_containers:
    description:
      - Information about the disk containers.
    type: list
    elements: dict
    suboptions:
      description:
        description:
          - The description of the disk image.
        type: str
      device_name:
        description:
          - The block device mapping for the disk.
        type: str
      format:
        description:
          - The format of the disk image being imported.
        type: str
        choices: ["OVA", "ova", "VHD", "vhd", "VHDX", "vhdx", "VMDK", "vmdk", "RAW", "raw"]
      snapshot_id:
        description:
          - The ID of the EBS snapshot to be used for importing the snapshot.
        type: str
      url:
        description:
          - The URL to the Amazon S3-based disk image being imported.
            The URL can either be a https URL (https://..) or an Amazon S3 URL (s3://..).
        type: str
      user_bucket:
          description:
          - The S3 bucket for the disk image.
          type: dict
          suboptions:
            s3_bucket:
              description:
                - The name of the Amazon S3 bucket where the disk image is located.
              type: str
            s3_key:
              description:
                - The file name of the disk image.
              type: str
  encrypted:
    description:
      - Specifies whether the destination AMI of the imported image should be encrypted.
      - The default KMS key for EBS is used unless you specify a non-default KMS key using O(kms_key_id).
    type: bool
  hypervisor:
    description:
      - The target hypervisor platform.
    type: str
    choices: ["xen"]
  kms_key_id:
    description:
      - An identifier for the symmetric KMS key to use when creating the encrypted AMI.
        This parameter is only required if you want to use a non-default KMS key;
        if this parameter is not specified, the default KMS key for EBS is used.
        If a O(kms_key_id) is specified, the O(encrypted) flag must also be set.
    type: str
  license_type:
    description:
      - The license type to be used for the Amazon Machine Image (AMI) after importing.
    type: str
  platform:
    description:
      - The operating system of the virtual machine.
    type: str
    choices: ["Windows", "Linux"]
  role_name:
    description:
      - The name of the role to use when not using the default role, 'vmimport'.
    type: str
  license_specifications:
    description:
      - The ARNs of the license configurations.
    type: list
    elements: dict
    suboptions:
      license_configuration_arn:
        description:
          - The ARN of a license configuration.
        type: str
  boot_mode:
    description:
      - The boot mode of the virtual machine.
    type: str
    choices: ["legacy-bios", "uefi"]
  cancel_reason:
    description:
      - The reason for canceling the task.
    type: str
  usage_operation:
    description:
      - The usage operation value.
    type: str
  tags:
    description:
      - The tags to apply to the import image task during creation.
    type: dict
    aliases: ["resource_tags"]
author:
  - Alina Buzachis (@alinabuzachis)
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.
- name: Import image
  amazon.aws.ec2_import_image:
    state: present
    task_name: "clone-vm-import-image"
    disk_containers:
      - format: raw
        user_bucket:
          s3_bucket: "clone-vm-s3-bucket"
          s3_key: "clone-vm-s3-bucket/ubuntu-vm-clone.raw"

- name: Cancel an import image task
  amazon.aws.ec2_import_image:
    state: absent
    task_name: "clone-vm-import-image"
"""

RETURN = r"""
import_image:
  description: A dict containing information about an EC2 import task.
  returned: always
  type: complex
  contains:
    task_name:
      description:
        - The name of the EC2 image import task.
      type: str
    architecture:
      description:
        - The architecture of the virtual machine.
      type: str
    image_id:
      description:
        - The ID of the Amazon Machine Image (AMI) created by the import task.
      type: str
    import_task_id:
      description:
        - The task ID of the import image task.
      type: str
    progress:
      description:
        - The progress of the task.
      type: str
    snapshot_details:
      description:
        - Describes the snapshot created from the imported disk.
      type: dict
      contains:
        description:
          description:
            - A description for the snapshot.
          type: str
        device_name:
          description:
            - The block device mapping for the snapshot.
          type: str
        disk_image_size:
          description:
            - The size of the disk in the snapshot, in GiB.
          type: float
        format:
          description:
            - The format of the disk image from which the snapshot is created.
          type: str
        progress:
          description:
            - The percentage of progress for the task.
          type: str
        snapshot_id:
          description:
            - The snapshot ID of the disk being imported.
          type: str
        status:
          description:
            - A brief status of the snapshot creation.
          type: str
        status_message:
          description:
            - A detailed status message for the snapshot creation.
          type: str
        url:
          description:
            - The URL used to access the disk image.
          type: str
        user_bucket:
          description:
            - The Amazon S3 bucket for the disk image.
          type: dict
    status:
      description:
        - A brief status of the task.
      type: str
    status_message:
      description:
        - A detailed status message of the import task.
      type: str
    license_specifications:
      description:
        - The ARNs of the license configurations.
      type: dict
    usage_operation:
      description:
        - The usage operation value.
      type: dict
    description:
      description:
        - A description string for the import image task.
      type: str
    encrypted:
      description:
        - Specifies whether the destination AMI of the imported image should be encrypted.
      type: bool
    hypervisor:
      description:
        - The target hypervisor platform.
      type: str
    kms_key_id:
      description:
        - The identifier for the symmetric KMS key that was used to create the encrypted AMI.
      type: str
    license_type:
      description:
        - The license type to be used for the Amazon Machine Image (AMI) after importing.
      type: str
    platform:
      description:
        - The operating system of the virtual machine.
      type: str
    role_name:
      description:
        - The name of the role to use when not using the default role, 'vmimport'.
      type: str
    tags:
      description:
        - The tags to apply to the import image task during creation.
      type: dict
"""

from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import cancel_import_task
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_import_image_tasks_as_snake_dict
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import import_image
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.transformation import scrub_none_parameters


def absent(client, module: AnsibleAWSModule) -> None:
    """
    Cancel an in-process import virtual machine
    """

    filters = {
        "Filters": [
            {"Name": "tag:Name", "Values": [module.params["task_name"]]},
            {"Name": "task-state", "Values": ["active"]},
        ]
    }

    try:
        import_image_info = describe_import_image_tasks_as_snake_dict(client, **filters)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)

    if import_image_info:
        result = import_image_info[0]
        result["task_name"] = module.params["task_name"]

        if module.check_mode:
            module.exit_json(changed=True, msg="Would have cancelled the import task if not in check mode")

        try:
            changed = cancel_import_task(
                client,
                import_task_id=result["import_task_id"],
                cancel_reason=module.params.get("cancel_reason"),
            )
            module.exit_json(changed=changed, import_image=result)
        except AnsibleEC2Error as e:
            module.fail_json_aws_error(e)
    else:
        module.exit_json(
            changed=False,
            msg="The specified import task does not exist or it cannot be cancelled",
            import_image={},
        )


def present(client, module: AnsibleAWSModule) -> None:
    params = {}
    tags = module.params.get("tags") or {}
    tags.update({"Name": module.params["task_name"]})

    if module.params.get("architecture"):
        params["Architecture"] = module.params["architecture"]
    if module.params.get("client_data"):
        params["ClientData"] = snake_dict_to_camel_dict(module.params["client_data"], capitalize_first=True)
    if module.params.get("description"):
        params["Description"] = module.params["description"]
    if module.params.get("disk_containers"):
        params["DiskContainers"] = snake_dict_to_camel_dict(module.params["disk_containers"], capitalize_first=True)
    if module.params.get("encrypted"):
        params["Encrypted"] = module.params["encrypted"]
    if module.params.get("hypervisor"):
        params["Hypervisor"] = module.params["hypervisor"]
    if module.params.get("kms_key_id"):
        params["KmsKeyId"] = module.params["kms_key_id"]
    if module.params.get("license_type"):
        params["LicenseType"] = module.params["license_type"]
    if module.params.get("platform"):
        params["Platform"] = module.params["platform"]
    if module.params.get("role_name"):
        params["RoleName"] = module.params["role_name"]
    if module.params.get("license_specifications"):
        params["LicenseSpecifications"] = snake_dict_to_camel_dict(
            module.params["license_specifications"], capitalize_first=True
        )
    if module.params.get("usage_operation"):
        params["UsageOperation"] = module.params["usage_operation"]
    if module.params.get("boot_mode"):
        params["BootMode"] = module.params.get("boot_mode")
    params["TagSpecifications"] = boto3_tag_specifications(tags, ["import-image-task"])

    filters = {
        "Filters": [
            {"Name": "tag:Name", "Values": [module.params["task_name"]]},
            {"Name": "task-state", "Values": ["completed", "active", "deleting"]},
        ]
    }
    try:
        import_image_info = describe_import_image_tasks_as_snake_dict(client, **filters)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)

    if import_image_info:
        result = import_image_info[0]
        result["task_name"] = module.params["task_name"]
        module.exit_json(
            changed=False,
            msg="An import task with the specified name already exists",
            import_image=result,
        )
    else:
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have created the import task if not in check mode")

        params = scrub_none_parameters(params)

        try:
            import_image(client, **params)
            import_image_info = describe_import_image_tasks_as_snake_dict(client, **filters)
            result = {}
            if import_image_info:
                result = import_image_info[0]
                result["task_name"] = module.params["task_name"]
            module.exit_json(
                changed=True,
                import_image=result,
            )
        except AnsibleEC2Error as e:
            module.fail_json_aws_error(e)


def main():
    argument_spec = dict(
        architecture=dict(type="str", choices=["i386", "x86_64"]),
        client_data=dict(
            type="dict",
            options=dict(
                comment=dict(type="str"),
                upload_end=dict(type="str"),
                upload_size=dict(type="float"),
                upload_start=dict(type="str"),
            ),
        ),
        description=dict(type="str"),
        license_specifications=dict(
            type="list",
            elements="dict",
            options=dict(
                license_configuration_arn=dict(type="str"),
            ),
        ),
        encrypted=dict(type="bool"),
        state=dict(default="present", choices=["present", "absent"]),
        hypervisor=dict(type="str", choices=["xen"]),
        kms_key_id=dict(type="str"),
        license_type=dict(type="str", no_log=False),
        tags=dict(required=False, type="dict", aliases=["resource_tags"]),
        platform=dict(type="str", choices=["Windows", "Linux"]),
        role_name=dict(type="str"),
        disk_containers=dict(
            type="list",
            elements="dict",
            options=dict(
                description=dict(type="str"),
                device_name=dict(type="str"),
                format=dict(
                    type="str", choices=["OVA", "ova", "VHD", "vhd", "VHDX", "vhdx", "VMDK", "vmdk", "RAW", "raw"]
                ),
                snapshot_id=dict(type="str"),
                url=dict(type="str"),
                user_bucket=dict(
                    type="dict",
                    options=dict(
                        s3_bucket=dict(type="str"),
                        s3_key=dict(type="str", no_log=True),
                    ),
                ),
            ),
        ),
        usage_operation=dict(type="str"),
        boot_mode=dict(type="str", choices=["legacy-bios", "uefi"]),
        cancel_reason=dict(type="str"),
        task_name=dict(type="str", aliases=["name"], required=True),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    state = module.params.get("state")
    client = module.client("ec2")

    if state == "present":
        present(client, module)
    else:
        absent(client, module)


if __name__ == "__main__":
    main()



# Copyright (c) 2017, Edward Nunez <edward.nunez@cyberark.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: cyberarkpassword
short_description: Get secrets from CyberArk AIM
requirements:
  - CyberArk AIM tool installed
description:
  - Get secrets from CyberArk AIM.
options:
  _command:
    description: Cyberark CLI utility.
    type: string
    env:
      - name: AIM_CLIPASSWORDSDK_CMD
    default: '/opt/CARKaim/sdk/clipasswordsdk'
  appid:
    description: Defines the unique ID of the application that is issuing the password request.
    type: string
    required: true
  query:
    description: Describes the filter criteria for the password retrieval.
    type: string
    required: true
  output:
    description:
      - Specifies the desired output fields separated by commas.
      - 'They could be: Password, PassProps.<property>, PasswordChangeInProcess.'
    type: string
    default: 'password'
  _extra:
    description: For extra_params values please check parameters for clipasswordsdk in CyberArk's "Credential Provider and
      ASCP Implementation Guide".
notes:
  - For Ansible on Windows, please change the -parameters (C(-p), C(-d), and C(-o)) to /parameters (C(/p), C(/d), and C(/o))
    and change the location of C(CLIPasswordSDK.exe).
"""

EXAMPLES = r"""
- name: passing options to the lookup
  ansible.builtin.debug:
    msg: '{{ lookup("community.general.cyberarkpassword", cyquery) }}'
  vars:
    cyquery:
      appid: "app_ansible"
      query: "safe=CyberArk_Passwords;folder=root;object=AdminPass"
      output: "Password,PassProps.UserName,PassProps.Address,PasswordChangeInProcess"


- name: used in a loop
  ansible.builtin.debug:
    msg: "{{item}}"
  with_community.general.cyberarkpassword:
    appid: 'app_ansible'
    query: 'safe=CyberArk_Passwords;folder=root;object=AdminPass'
    output: 'Password,PassProps.UserName,PassProps.Address,PasswordChangeInProcess'
"""

RETURN = r"""
_result:
  description: A list containing one dictionary.
  type: list
  elements: dictionary
  contains:
    password:
      description:
        - The actual value stored.
    passprops:
      description: Properties assigned to the entry.
      type: dictionary
    passwordchangeinprocess:
      description: Did the password change?
"""

import os
import subprocess
from subprocess import PIPE, Popen

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_bytes, to_native
from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display

display = Display()

CLIPASSWORDSDK_CMD = os.getenv("AIM_CLIPASSWORDSDK_CMD", "/opt/CARKaim/sdk/clipasswordsdk")


class CyberarkPassword:
    def __init__(self, appid=None, query=None, output=None, **kwargs):
        self.appid = appid
        self.query = query
        self.output = output

        # Support for Generic parameters to be able to specify
        # FailRequestOnPasswordChange, Queryformat, Reason, etc.
        self.extra_parms = []
        for key, value in kwargs.items():
            self.extra_parms.append("-p")
            self.extra_parms.append(f"{key}={value}")

        if self.appid is None:
            raise AnsibleError("CyberArk Error: No Application ID specified")
        if self.query is None:
            raise AnsibleError("CyberArk Error: No Vault query specified")

        if self.output is None:
            # If no output is specified, return at least the password
            self.output = "password"
        else:
            # To avoid reference issues/confusion to values, all
            # output 'keys' will be in lowercase.
            self.output = self.output.lower()

        self.b_delimiter = b"@#@"  # Known delimiter to split output results

    def get(self):
        result_dict = {}

        try:
            all_parms = [
                CLIPASSWORDSDK_CMD,
                "GetPassword",
                "-p",
                f"AppDescs.AppID={self.appid}",
                "-p",
                f"Query={self.query}",
                "-o",
                self.output,
                "-d",
                self.b_delimiter,
            ]
            all_parms.extend(self.extra_parms)

            b_credential = b""
            b_all_params = [to_bytes(v) for v in all_parms]
            tmp_output, tmp_error = Popen(b_all_params, stdout=PIPE, stderr=PIPE, stdin=PIPE).communicate()

            if tmp_output:
                b_credential = to_bytes(tmp_output)

            if tmp_error:
                raise AnsibleError(f"ERROR => {tmp_error} ")

            if b_credential and b_credential.endswith(b"\n"):
                b_credential = b_credential[:-1]

            output_names = self.output.split(",")
            output_values = b_credential.split(self.b_delimiter)

            for i in range(len(output_names)):
                if output_names[i].startswith("passprops."):
                    if "passprops" not in result_dict:
                        result_dict["passprops"] = {}
                    output_prop_name = output_names[i][10:]
                    result_dict["passprops"][output_prop_name] = to_native(output_values[i])
                else:
                    result_dict[output_names[i]] = to_native(output_values[i])

        except subprocess.CalledProcessError as e:
            raise AnsibleError(e.output) from e
        except OSError as e:
            raise AnsibleError(
                f"ERROR - AIM not installed or clipasswordsdk not in standard location. ERROR=({e.errno}) => {e.strerror} "
            ) from e

        return [result_dict]


class LookupModule(LookupBase):
    """
    USAGE:

    """

    def run(self, terms, variables=None, **kwargs):
        # TODO: use new-style option parsing
        display.vvvv(f"{terms}")
        if isinstance(terms, list):
            return_values = []
            for term in terms:
                display.vvvv(f"Term: {term}")
                cyberark_conn = CyberarkPassword(**term)
                return_values.append(cyberark_conn.get())
            return return_values
        else:
            cyberark_conn = CyberarkPassword(**terms)
            result = cyberark_conn.get()
            return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}


DOCUMENTATION = r"""
---
module: aci_esg_epg_selector
short_description: Manage ESG - EPG Selectors (fv:EPgSelector)
description:
- Manage Endpoint Security Groups - EPG Selectors on Cisco ACI fabrics.

options:
  tenant:
    description:
    - Name of the tenant.
    type: str
    aliases: [ tenant_name ]
  ap:
    description:
    - The name of the application profile.
    type: str
    aliases: [ app_profile, app_profile_name ]
  esg:
    description:
    - Name of the Endpoint Security Group.
    type: str
    aliases: [ esg_name ]
  epg_ap:
    description:
    - Name of the Application profile which contains the EPG.
    type: str
  epg:
    description:
    - Name of the EPG which is used to create EPG Selector object under the ESG.
    type: str
    aliases: [ epg_name ]
  description:
    description:
    - Description of the ESG Tag Selector.
    type: str
    aliases: [ epg_selector_description ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

seealso:
- module: cisco.aci.aci_esg
- name: Manage Endpoint Security Groups (ESGs) objects (fv:ESg)
  description: Manage Endpoint Security Groups (ESGs) on Cisco ACI fabrics.
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Sabari Jaganathan (@sajagana)
"""


EXAMPLES = r"""
- name: Add an EPG selector
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: production_ap
    esg: web_esg
    epg_ap: production_ap1
    epg: production_ap1-epg
    description: epg-test-description
    state: present
  delegate_to: localhost

- name: Add list of EPG selectors
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: production_ap
    esg: "{{ item.esg }}"
    epg_ap: "{{ item.epg_ap }}"
    epg: "{{ item.epg }}"
    description: epg-test-description
    state: present
  delegate_to: localhost
  with_items:
    - {"epg_ap": "production_ap1", "epg": "epg-test1", "esg": "web_esg"}
    - {"epg_ap": "production_ap1", "epg": "epg-test2", "esg": "web_esg"}

- name: Query an EPG selector with esg and epg name
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: production_ap
    esg: web_esg
    epg_ap: production_ap1
    epg: production_ap1-epg
    state: query
  delegate_to: localhost

- name: Query all EPG selectors under a application profile
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: production_ap
    state: query
  delegate_to: localhost

- name: Query all EPG selectors
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost

- name: Remove an EPG selector
  cisco.aci.aci_esg_epg_selector:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: production_ap
    esg: web_esg
    epg_ap: production_ap1
    epg: production_ap1-epg
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),
        ap=dict(type="str", aliases=["app_profile", "app_profile_name"]),
        esg=dict(type="str", aliases=["esg_name"]),
        epg_ap=dict(type="str"),
        epg=dict(type="str", aliases=["epg_name"]),
        description=dict(type="str", aliases=["epg_selector_description"]),
        state=dict(
            type="str",
            default="present",
            choices=["absent", "present", "query"],
        ),
        name_alias=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["tenant", "ap", "esg", "epg", "epg_ap"]],
            ["state", "present", ["tenant", "ap", "esg", "epg", "epg_ap"]],
        ],
    )

    aci = ACIModule(module)
    tenant = module.params.get("tenant")
    ap = module.params.get("ap")
    esg = module.params.get("esg")
    epg_ap = module.params.get("epg_ap")
    epg = module.params.get("epg")
    description = module.params.get("description")
    state = module.params.get("state")

    matchEpgDn = "uni/tn-{0}/ap-{1}/epg-{2}".format(tenant, epg_ap, epg)
    epgselector = "epgselector-[{0}]".format(matchEpgDn)
    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="fvAp",
            aci_rn="ap-{0}".format(ap),
            module_object=ap,
            target_filter={"name": ap},
        ),
        subclass_2=dict(
            aci_class="fvESg",
            aci_rn="esg-{0}".format(esg),
            module_object=esg,
            target_filter={"name": esg},
        ),
        subclass_3=dict(
            aci_class="fvEPgSelector",
            aci_rn=epgselector,
            module_object=epg,
            target_filter={"matchEpgDn": epg},
        ),
    )
    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="fvEPgSelector",
            class_config=dict(
                matchEpgDn=matchEpgDn,
                descr=description,
            ),
        )

        aci.get_diff(aci_class="fvEPgSelector")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Jan-Piet Mens <jpmens () gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: irc
short_description: Send a message to an IRC channel or a nick
description:
  - Send a message to an IRC channel or a nick. This is a very simplistic implementation.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  server:
    type: str
    description:
      - IRC server name/address.
    default: localhost
  port:
    type: int
    description:
      - IRC server port number.
    default: 6667
  nick:
    type: str
    description:
      - Nickname to send the message from. May be shortened, depending on server's NICKLEN setting.
    default: ansible
  msg:
    type: str
    description:
      - The message body.
    required: true
  topic:
    type: str
    description:
      - Set the channel topic.
  color:
    type: str
    description:
      - Text color for the message.
    default: "none"
    choices:
      - none
      - white
      - black
      - blue
      - green
      - red
      - brown
      - purple
      - orange
      - yellow
      - light_green
      - teal
      - light_cyan
      - light_blue
      - pink
      - gray
      - light_gray
    aliases: [colour]
  channel:
    type: str
    description:
      - Channel name. One of O(nick_to) or O(channel) needs to be set. When both are set, the message is sent to both of them.
  nick_to:
    type: list
    elements: str
    description:
      - A list of nicknames to send the message to. One of O(nick_to) or O(channel) needs to be set. When both are defined, the
        message is sent to both of them.
  key:
    type: str
    description:
      - Channel key.
  passwd:
    type: str
    description:
      - Server password.
  timeout:
    type: int
    description:
      - Timeout to use while waiting for successful registration and join messages, this is to prevent an endless loop.
    default: 30
  use_tls:
    description:
      - Designates whether TLS/SSL should be used when connecting to the IRC server.
      - O(use_tls) is available since community.general 8.1.0, before the option was exclusively called O(use_ssl). The latter
        is now an alias of O(use_tls).
      - B(Note:) for security reasons, you should always set O(use_tls=true) and O(validate_certs=true) whenever possible.
      - The default of this option changed to V(true) in community.general 10.0.0.
    type: bool
    default: true
    aliases:
      - use_ssl
  part:
    description:
      - Designates whether user should part from channel after sending message or not. Useful for when using a mock bot and
        not wanting join/parts between messages.
    type: bool
    default: true
  style:
    type: str
    description:
      - Text style for the message. Note italic does not work on some clients.
    choices: ["bold", "underline", "reverse", "italic", "none"]
    default: none
  validate_certs:
    description:
      - If set to V(false), the SSL certificates are not validated.
      - This should always be set to V(true). Using V(false) is unsafe and should only be done if the network between between
        Ansible and the IRC server is known to be safe.
      - B(Note:) for security reasons, you should always set O(use_tls=true) and O(validate_certs=true) whenever possible.
      - The default of this option changed to V(true) in community.general 10.0.0.
    type: bool
    default: true
    version_added: 8.1.0

# informational: requirements for nodes
requirements: [socket]
author:
  - "Jan-Piet Mens (@jpmens)"
  - "Matt Martz (@sivel)"
"""

EXAMPLES = r"""
- name: Send a message to an IRC channel from nick ansible
  community.general.irc:
    server: irc.example.net
    use_tls: true
    validate_certs: true
    channel: '#t1'
    msg: Hello world

- name: Send a message to an IRC channel
  local_action:
    module: irc
    port: 6669
    server: irc.example.net
    use_tls: true
    validate_certs: true
    channel: '#t1'
    msg: 'All finished at {{ ansible_date_time.iso8601 }}'
    color: red
    nick: ansibleIRC

- name: Send a message to an IRC channel
  local_action:
    module: irc
    port: 6669
    server: irc.example.net
    use_tls: true
    validate_certs: true
    channel: '#t1'
    nick_to:
      - nick1
      - nick2
    msg: 'All finished at {{ ansible_date_time.iso8601 }}'
    color: red
    nick: ansibleIRC
"""

# ===========================================
# IRC module support methods.
#

import re
import socket
import ssl
import time
import traceback

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_native


def send_msg(
    msg,
    server="localhost",
    port="6667",
    channel=None,
    nick_to=None,
    key=None,
    topic=None,
    nick="ansible",
    color="none",
    passwd=False,
    timeout=30,
    use_tls=False,
    validate_certs=True,
    part=True,
    style=None,
):
    """send message to IRC"""
    nick_to = [] if nick_to is None else nick_to

    colornumbers = {
        "white": "00",
        "black": "01",
        "blue": "02",
        "green": "03",
        "red": "04",
        "brown": "05",
        "purple": "06",
        "orange": "07",
        "yellow": "08",
        "light_green": "09",
        "teal": "10",
        "light_cyan": "11",
        "light_blue": "12",
        "pink": "13",
        "gray": "14",
        "light_gray": "15",
    }

    stylechoices = {
        "bold": "\x02",
        "underline": "\x1f",
        "reverse": "\x16",
        "italic": "\x1d",
    }

    try:
        styletext = stylechoices[style]
    except Exception:
        styletext = ""

    try:
        colornumber = colornumbers[color]
        colortext = f"\x03{colornumber}"
    except Exception:
        colortext = ""

    message = styletext + colortext + msg

    irc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if use_tls:
        kwargs = {}
        if validate_certs:
            context = ssl.create_default_context()
            kwargs["server_hostname"] = server
        else:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS)
            context.verify_mode = ssl.CERT_NONE
        irc = context.wrap_socket(irc, **kwargs)
    irc.connect((server, int(port)))

    if passwd:
        irc.send(to_bytes(f"PASS {passwd}\r\n"))
    irc.send(to_bytes(f"NICK {nick}\r\n"))
    irc.send(to_bytes(f"USER {nick} {nick} {nick} :ansible IRC\r\n"))
    motd = ""
    start = time.time()
    while True:
        motd += to_native(irc.recv(1024))
        # The server might send back a shorter nick than we specified (due to NICKLEN),
        #  so grab that and use it from now on (assuming we find the 00[1-4] response).
        match = re.search(r"^:\S+ 00[1-4] (?P<nick>\S+) :", motd, flags=re.M)
        if match:
            nick = match.group("nick")
            break
        elif time.time() - start > timeout:
            raise Exception("Timeout waiting for IRC server welcome response")
        time.sleep(0.5)

    if channel:
        if key:
            irc.send(to_bytes(f"JOIN {channel} {key}\r\n"))
        else:
            irc.send(to_bytes(f"JOIN {channel}\r\n"))

        join = ""
        start = time.time()
        while True:
            join += to_native(irc.recv(1024))
            if re.search(rf"^:\S+ 366 {nick} {channel} :", join, flags=re.M | re.I):
                break
            elif time.time() - start > timeout:
                raise Exception("Timeout waiting for IRC JOIN response")
            time.sleep(0.5)

        if topic is not None:
            irc.send(to_bytes(f"TOPIC {channel} :{topic}\r\n"))
            time.sleep(1)

    if nick_to:
        for nick in nick_to:
            irc.send(to_bytes(f"PRIVMSG {nick} :{message}\r\n"))
    if channel:
        irc.send(to_bytes(f"PRIVMSG {channel} :{message}\r\n"))
    time.sleep(1)
    if part:
        if channel:
            irc.send(to_bytes(f"PART {channel}\r\n"))
        irc.send(to_bytes("QUIT\r\n"))
        time.sleep(1)
    irc.close()


# ===========================================
# Main
#


def main():
    module = AnsibleModule(
        argument_spec=dict(
            server=dict(default="localhost"),
            port=dict(type="int", default=6667),
            nick=dict(default="ansible"),
            nick_to=dict(type="list", elements="str"),
            msg=dict(required=True),
            color=dict(
                default="none",
                aliases=["colour"],
                choices=[
                    "white",
                    "black",
                    "blue",
                    "green",
                    "red",
                    "brown",
                    "purple",
                    "orange",
                    "yellow",
                    "light_green",
                    "teal",
                    "light_cyan",
                    "light_blue",
                    "pink",
                    "gray",
                    "light_gray",
                    "none",
                ],
            ),
            style=dict(default="none", choices=["underline", "reverse", "bold", "italic", "none"]),
            channel=dict(),
            key=dict(no_log=True),
            topic=dict(),
            passwd=dict(no_log=True),
            timeout=dict(type="int", default=30),
            part=dict(type="bool", default=True),
            use_tls=dict(type="bool", default=True, aliases=["use_ssl"]),
            validate_certs=dict(type="bool", default=True),
        ),
        supports_check_mode=True,
        required_one_of=[["channel", "nick_to"]],
    )

    server = module.params["server"]
    port = module.params["port"]
    nick = module.params["nick"]
    nick_to = module.params["nick_to"]
    msg = module.params["msg"]
    color = module.params["color"]
    channel = module.params["channel"]
    topic = module.params["topic"]
    if topic and not channel:
        module.fail_json(msg="When topic is specified, a channel is required.")
    key = module.params["key"]
    passwd = module.params["passwd"]
    timeout = module.params["timeout"]
    use_tls = module.params["use_tls"]
    part = module.params["part"]
    style = module.params["style"]
    validate_certs = module.params["validate_certs"]

    try:
        send_msg(
            msg,
            server,
            port,
            channel,
            nick_to,
            key,
            topic,
            nick,
            color,
            passwd,
            timeout,
            use_tls,
            validate_certs,
            part,
            style,
        )
    except Exception as e:
        module.fail_json(msg=f"unable to send to IRC: {e}", exception=traceback.format_exc())

    module.exit_json(changed=False, channel=channel, nick=nick, msg=msg)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hetzner_dns_zone_info

short_description: Retrieve zone information in Hetzner DNS service

version_added: 2.0.0

description:
  - Retrieves zone information in Hetzner DNS service.
extends_documentation_fragment:
  - community.dns._hetzner
  - community.dns._hetzner.zone_id_type
  - community.dns._module_zone_info
  - community.dns._attributes
  - community.dns._attributes.actiongroup_hetzner
  - community.dns._attributes.info_module
  - community.dns._attributes.idempotent_not_modify_state

attributes:
  action_group:
    version_added: 2.4.0

author:
  - Markus Bergholz (@markuman) <markuman+spambelongstogoogle@gmail.com>
  - Felix Fontein (@felixfontein)
"""

EXAMPLES = r"""
- name: Retrieve details for foo.com zone
  community.dns.hetzner_dns_zone_info:
    zone: foo.com
    hetzner_token: access_token
  register: rec

- name: Retrieve details for zone 23
  community.dns.hetzner_dns_zone_info:
    zone_id: 23
    hetzner_token: access_token
"""

RETURN = r"""
zone_name:
  description: The name of the zone.
  type: int
  returned: success
  sample: example.com

zone_id:
  description: The ID of the zone.
  type: str
  returned: success
  sample: 23

zone_info:
  description:
    - Extra information returned by the API.
  type: dict
  returned: success
  contains:
    # TODO: adjust to new API!
    created:
      description:
        - The time when the zone was created.
      type: str
      sample: "2021-07-15T19:23:58Z"
    modified:
      description:
        - The time the zone was last modified.
      type: str
      sample: "2021-07-15T19:23:58Z"
    legacy_dns_host:
      description:
        # TODO
        - Unknown.
      type: str
      returned: if zone was imported
    legacy_ns:
      description:
        - List of nameservers during import.
      type: list
      elements: str
      returned: if zone was imported
    ns:
      description:
        - List of nameservers the zone should have for using Hetzner's DNS.
      type: list
      elements: str
    owner:
      description:
        - Owner of the zone.
      type: str
    paused:
      description:
        # TODO
        - Unknown.
      type: bool
      sample: true
    permission:
      description:
        - Zone's permissions.
      type: str
    project:
      description:
        # TODO
        - Unknown.
      type: str
    registrar:
      description:
        # TODO
        - Unknown.
      type: str
    status:
      description:
        - Status of the zone.
        - Can be one of V(verified), V(failed) and V(pending).
      type: str
      sample: verified
      # TODO
      # choices:
      #   - verified
      #   - failed
      #   - pending
    ttl:
      description:
        - TTL of zone.
      type: int
      sample: 0
    verified:
      description:
        - Time when zone was verified.
      type: str
      sample: "2021-07-15T19:23:58Z"
    records_count:
      description:
        - Number of records associated to this zone.
      type: int
      sample: 0
    is_secondary_dns:
      description:
        - Indicates whether the zone is a secondary DNS zone.
      type: bool
      sample: true
    txt_verification:
      description:
        - Shape of the TXT record that has to be set to verify a zone.
        - If name and token are empty, no TXT record needs to be set.
      type: dict
      sample: {'name': '', 'token': ''}
      contains:
        name:
          description:
            - The TXT record's name.
          type: str
        token:
          description:
            - The TXT record's content.
          type: str
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.dns.plugins.module_utils._argspec import (
    ModuleOptionProvider,
)
from ansible_collections.community.dns.plugins.module_utils._hetzner.api import (
    create_hetzner_api,
    create_hetzner_argument_spec,
    create_hetzner_provider_information,
)
from ansible_collections.community.dns.plugins.module_utils._http import (
    ModuleHTTPHelper,
)
from ansible_collections.community.dns.plugins.module_utils._module.zone_info import (
    create_module_argument_spec,
    run_module,
)


def main() -> None:
    provider_information = create_hetzner_provider_information()
    argument_spec = create_hetzner_argument_spec()
    argument_spec.merge(
        create_module_argument_spec(provider_information=provider_information)
    )
    module = AnsibleModule(supports_check_mode=True, **argument_spec.to_kwargs())
    option_provider = ModuleOptionProvider(module)
    run_module(
        module,
        lambda: create_hetzner_api(option_provider, ModuleHTTPHelper(module)),
        provider_information=create_hetzner_provider_information(
            option_provider=option_provider
        ),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2021, Raphaël Droz (raphael.droz@gmail.com)
# Copyright (c) 2019, Guillaume Martinez (lunik@tiwabbit.fr)
# Copyright (c) 2018, Samy Coenen <samy.coenen@nubera.be>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_runner
short_description: Create, modify and delete GitLab Runners
description:
  - Register, update and delete runners on GitLab Server side with the GitLab API.
  - All operations are performed using the GitLab API v4.
  - For details, consult the full API documentation at U(https://docs.gitlab.com/ee/api/runners.html) and
    U(https://docs.gitlab.com/ee/api/users.html#create-a-runner-linked-to-a-user).
  - A valid private API token is required for all operations. You can create as many tokens as you like using the GitLab web
    interface at U(https://$GITLAB_URL/profile/personal_access_tokens).
  - A valid registration token is required for registering a new runner. To create shared runners, you need to ask your administrator
    to give you this token. It can be found at U(https://$GITLAB_URL/admin/runners/).
  - This module does not handle the C(gitlab-runner) process part, but only manages the runner on GitLab Server side through
    its API. Once the module has created the runner, you may use the generated token to run C(gitlab-runner register) command.
notes:
  - To create a new runner at least the O(api_token), O(description) and O(api_url) options are required.
  - Runners need to have unique descriptions, since this attribute is used as key for idempotency.
author:
  - Samy Coenen (@SamyCoenen)
  - Guillaume Martinez (@Lunik)
requirements:
  - python-gitlab >= 1.5.0 for legacy runner registration workflow (runner registration token -
    U(https://docs.gitlab.com/runner/register/#register-with-a-runner-registration-token-deprecated))
  - python-gitlab >= 4.0.0 for new runner registration workflow (runner authentication token -
    U(https://docs.gitlab.com/runner/register/#register-with-a-runner-authentication-token))
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  group:
    description:
      - ID or full path of the group in the form group/subgroup.
      - Mutually exclusive with O(owned) and O(project).
      - Must be group's numeric ID if O(registration_token) is not set and O(state=present).
    type: str
    version_added: '6.5.0'
  project:
    description:
      - ID or full path of the project in the form of group/name.
      - Mutually exclusive with O(owned) since community.general 4.5.0.
      - Mutually exclusive with O(group).
      - Must be project's numeric ID if O(registration_token) is not set and O(state=present).
    type: str
    version_added: '3.7.0'
  description:
    description:
      - The unique name of the runner.
    required: true
    type: str
    aliases:
      - name
  state:
    description:
      - Make sure that the runner with the same name exists with the same configuration or delete the runner with the same
        name.
    default: present
    choices: ["present", "absent"]
    type: str
  registration_token:
    description:
      - The registration token is used to register new runners before GitLab 16.0.
      - Required if O(state=present) for GitLab < 16.0.
      - If set, the runner is created using the old runner creation workflow.
      - If not set, the runner is created using the new runner creation workflow, introduced in GitLab 16.0.
      - If not set, requires python-gitlab >= 4.0.0.
    type: str
  owned:
    description:
      - Searches only runners available to the user when searching for existing, when false admin token required.
      - Mutually exclusive with O(project) since community.general 4.5.0.
      - Mutually exclusive with O(group).
    default: false
    type: bool
    version_added: 2.0.0
  active:
    description:
      - Define if the runners is immediately active after creation.
      - Mutually exclusive with O(paused).
    default: true
    type: bool
  paused:
    description:
      - Define if the runners is active or paused after creation.
      - Mutually exclusive with O(active).
    default: false
    type: bool
    version_added: 8.1.0
  locked:
    description:
      - Determines if the runner is locked or not.
    default: false
    type: bool
  access_level:
    description:
      - Determines if a runner can pick up jobs only from protected branches.
      - If O(access_level_on_creation) is not explicitly set to V(true), this option is ignored on registration and is only
        applied on updates.
      - If set to V(not_protected), runner can pick up jobs from both protected and unprotected branches.
      - If set to V(ref_protected), runner can pick up jobs only from protected branches.
      - Before community.general 8.0.0 the default was V(ref_protected). This was changed to no default in community.general
        8.0.0. If this option is not specified explicitly, GitLab uses V(not_protected) on creation, and the value set is
        not changed on any updates.
    choices: ["not_protected", "ref_protected"]
    type: str
  access_level_on_creation:
    description:
      - Whether the runner should be registered with an access level or not.
      - If set to V(true), the value of O(access_level) is used for runner registration.
      - If set to V(false), GitLab registers the runner with the default access level.
      - The default of this option changed to V(true) in community.general 7.0.0. Before, it was V(false).
    default: true
    type: bool
    version_added: 6.3.0
  maximum_timeout:
    description:
      - The maximum time that a runner has to complete a specific job.
      - Use V(0) to disable the timeout. This is available since community.general 12.1.0.
    default: 3600
    type: int
  run_untagged:
    description:
      - Run untagged jobs or not.
    default: true
    type: bool
  tag_list:
    description: The tags that apply to the runner.
    default: []
    type: list
    elements: str
"""

EXAMPLES = r"""
- name: Create an instance-level runner
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    description: Docker Machine t1
    state: present
    active: true
    tag_list: ['docker']
    run_untagged: false
    locked: false
  register: runner # Register module output to run C(gitlab-runner register) command in another task

- name: Create a group-level runner
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    description: Docker Machine t1
    state: present
    active: true
    tag_list: ['docker']
    run_untagged: false
    locked: false
    group: top-level-group/subgroup
  register: runner # Register module output to run C(gitlab-runner register) command in another task

- name: Create a project-level runner
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    description: Docker Machine t1
    state: present
    active: true
    tag_list: ['docker']
    run_untagged: false
    locked: false
    project: top-level-group/subgroup/project
  register: runner # Register module output to run C(gitlab-runner register) command in another task

- name: "Register instance-level runner with registration token (deprecated)"
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    registration_token: 4gfdsg345
    description: Docker Machine t1
    state: present
    active: true
    tag_list: ['docker']
    run_untagged: false
    locked: false
  register: runner # Register module output to run C(gitlab-runner register) command in another task

- name: "Delete runner"
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    description: Docker Machine t1
    state: absent

- name: Delete an owned runner as a non-admin
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    description: Docker Machine t1
    owned: true
    state: absent

- name: "Register a project-level runner with registration token (deprecated)"
  community.general.gitlab_runner:
    api_url: https://gitlab.example.com/
    api_token: "{{ access_token }}"
    registration_token: 4gfdsg345
    description: MyProject runner
    state: present
    project: mygroup/mysubgroup/myproject
  register: runner # Register module output to run C(gitlab-runner register) command in another task
"""

RETURN = r"""
msg:
  description: Success or failure message.
  returned: always
  type: str
  sample: "Success"

result:
  description: JSON-parsed response from the server.
  returned: always
  type: dict

error:
  description: The error message returned by the GitLab API.
  returned: failed
  type: str
  sample: "400: path is already in use"

runner:
  description: API object.
  returned: always
  type: dict
"""

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    gitlab,
    gitlab_authentication,
    list_all_kwargs,
)
from ansible_collections.community.general.plugins.module_utils._version import LooseVersion


class GitLabRunner:
    def __init__(self, module, gitlab_instance, group=None, project=None):
        self._module = module
        self._gitlab = gitlab_instance
        self.runner_object = None

        # Whether to operate on GitLab-instance-wide or project-wide runners
        # See https://gitlab.com/gitlab-org/gitlab-ce/issues/60774
        # for group runner token access
        if project:
            self._runners_endpoint = project.runners.list
        elif group:
            self._runners_endpoint = group.runners.list
        elif module.params["owned"]:
            self._runners_endpoint = gitlab_instance.runners.list
        else:
            self._runners_endpoint = gitlab_instance.runners.all

    def create_or_update_runner(self, description, options):
        changed = False

        arguments = {
            "locked": options["locked"],
            "run_untagged": options["run_untagged"],
            "maximum_timeout": options["maximum_timeout"],
            "tag_list": options["tag_list"],
        }

        if options.get("paused") is not None:
            arguments["paused"] = options["paused"]
        else:
            arguments["active"] = options["active"]

        if options.get("access_level") is not None:
            arguments["access_level"] = options["access_level"]
        # Because we have already call userExists in main()
        if self.runner_object is None:
            arguments["description"] = description
            if options.get("registration_token") is not None:
                arguments["token"] = options["registration_token"]
            elif options.get("group") is not None:
                arguments["runner_type"] = "group_type"
                arguments["group_id"] = options["group"]
            elif options.get("project") is not None:
                arguments["runner_type"] = "project_type"
                arguments["project_id"] = options["project"]
            else:
                arguments["runner_type"] = "instance_type"

            access_level_on_creation = self._module.params["access_level_on_creation"]
            if not access_level_on_creation:
                arguments.pop("access_level", None)

            runner = self.create_runner(arguments)
            changed = True
        else:
            changed, runner = self.update_runner(self.runner_object, arguments)
            if changed:
                if self._module.check_mode:
                    self._module.exit_json(changed=True, msg=f"Successfully updated the runner {description}")

                try:
                    runner.save()
                except Exception as e:
                    self._module.fail_json(msg=f"Failed to update runner: {e} ")

        self.runner_object = runner
        return changed

    """
    @param arguments Attributes of the runner
    """

    def create_runner(self, arguments):
        if self._module.check_mode:

            class MockRunner:
                def __init__(self):
                    self._attrs = {}

            return MockRunner()

        try:
            if arguments.get("token") is not None:
                runner = self._gitlab.runners.create(arguments)
            elif LooseVersion(gitlab.__version__) < LooseVersion("4.0.0"):
                self._module.fail_json(msg="New runner creation workflow requires python-gitlab 4.0.0 or higher")
            else:
                runner = self._gitlab.user.runners.create(arguments)
        except gitlab.exceptions.GitlabCreateError as e:
            self._module.fail_json(msg=f"Failed to create runner: {e}")

        return runner

    """
    @param runner Runner object
    @param arguments Attributes of the runner
    """

    def update_runner(self, runner, arguments):
        changed = False

        for arg_key, arg_value in arguments.items():
            if arg_value is not None or arg_key == "maximum_timeout":
                if isinstance(arg_value, list):
                    list1 = getattr(runner, arg_key)
                    list1.sort()
                    list2 = arg_value
                    list2.sort()
                    if list1 != list2:
                        setattr(runner, arg_key, arg_value)
                        changed = True
                else:
                    if getattr(runner, arg_key) != arg_value:
                        setattr(runner, arg_key, arg_value)
                        changed = True

        return (changed, runner)

    """
    @param description Description of the runner
    """

    def find_runner(self, description):
        runners = self._runners_endpoint(**list_all_kwargs)

        for runner in runners:
            # python-gitlab 2.2 through at least 2.5 returns a list of dicts for list() instead of a Runner
            # object, so we need to handle both
            if hasattr(runner, "description"):
                if runner.description == description:
                    return self._gitlab.runners.get(runner.id)
            else:
                if runner["description"] == description:
                    return self._gitlab.runners.get(runner["id"])

    """
    @param description Description of the runner
    """

    def exists_runner(self, description):
        # When runner exists, object will be stored in self.runner_object.
        runner = self.find_runner(description)

        if runner:
            self.runner_object = runner
            return True
        return False

    def delete_runner(self):
        if self._module.check_mode:
            return True

        runner = self.runner_object

        return runner.delete()


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        dict(
            description=dict(type="str", required=True, aliases=["name"]),
            active=dict(type="bool", default=True),
            paused=dict(type="bool", default=False),
            owned=dict(type="bool", default=False),
            tag_list=dict(type="list", elements="str", default=[]),
            run_untagged=dict(type="bool", default=True),
            locked=dict(type="bool", default=False),
            access_level=dict(type="str", choices=["not_protected", "ref_protected"]),
            access_level_on_creation=dict(type="bool", default=True),
            maximum_timeout=dict(type="int", default=3600),
            registration_token=dict(type="str", no_log=True),
            project=dict(type="str"),
            group=dict(type="str"),
            state=dict(type="str", default="present", choices=["absent", "present"]),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
            ["project", "owned"],
            ["group", "owned"],
            ["project", "group"],
            ["active", "paused"],
        ],
        required_together=[
            ["api_username", "api_password"],
        ],
        required_one_of=[
            ["api_username", "api_token", "api_oauth_token", "api_job_token"],
        ],
        supports_check_mode=True,
    )

    # check prerequisites and connect to gitlab server
    gitlab_instance = gitlab_authentication(module)

    state = module.params["state"]
    runner_description = module.params["description"]
    runner_active = module.params["active"]
    runner_paused = module.params["paused"]
    tag_list = module.params["tag_list"]
    run_untagged = module.params["run_untagged"]
    runner_locked = module.params["locked"]
    access_level = module.params["access_level"]
    if module.params["maximum_timeout"] == 0:
        maximum_timeout = None
    else:
        maximum_timeout = module.params["maximum_timeout"]
    registration_token = module.params["registration_token"]
    project = module.params["project"]
    group = module.params["group"]

    gitlab_project = None
    gitlab_group = None

    if project:
        try:
            gitlab_project = gitlab_instance.projects.get(project)
        except gitlab.exceptions.GitlabGetError as e:
            module.fail_json(msg=f"No such a project {project}", exception=e)
    elif group:
        try:
            gitlab_group = gitlab_instance.groups.get(group)
        except gitlab.exceptions.GitlabGetError as e:
            module.fail_json(msg=f"No such a group {group}", exception=e)

    gitlab_runner = GitLabRunner(module, gitlab_instance, gitlab_group, gitlab_project)
    runner_exists = gitlab_runner.exists_runner(runner_description)

    if state == "absent":
        if runner_exists:
            gitlab_runner.delete_runner()
            module.exit_json(changed=True, msg=f"Successfully deleted runner {runner_description}")
        else:
            module.exit_json(changed=False, msg="Runner deleted or does not exists")

    if state == "present":
        runner_values = {
            "active": runner_active,
            "tag_list": tag_list,
            "run_untagged": run_untagged,
            "locked": runner_locked,
            "access_level": access_level,
            "maximum_timeout": maximum_timeout,
            "registration_token": registration_token,
            "group": group,
            "project": project,
        }
        if LooseVersion(gitlab_runner._gitlab.version()[0]) >= LooseVersion("14.8.0"):
            # the paused attribute for runners is available since 14.8
            runner_values["paused"] = runner_paused
        if gitlab_runner.create_or_update_runner(runner_description, runner_values):
            module.exit_json(
                changed=True,
                runner=gitlab_runner.runner_object._attrs,
                msg=f"Successfully created or updated the runner {runner_description}",
            )
        else:
            module.exit_json(
                changed=False,
                runner=gitlab_runner.runner_object._attrs,
                msg=f"No need to update the runner {runner_description}",
            )


if __name__ == "__main__":
    main()



# Copyright (c) 2016-2017, Yanis Guenane <yanis+ansible@guenane.org>
# Copyright (c) 2017, Markus Teufelberger <mteufelberger+ansible@mgit.at>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import typing as t
from random import randrange

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLBadPassphraseError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_support import (
    CRYPTOGRAPHY_TIMEZONE,
    cryptography_compare_public_keys,
    cryptography_key_needs_digest_for_signing,
    cryptography_verify_certificate_signature,
    get_not_valid_after,
    get_not_valid_before,
    is_potential_certificate_issuer_public_key,
    set_not_valid_after,
    set_not_valid_before,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.certificate import (
    CertificateBackend,
    CertificateError,
    CertificateProvider,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.support import (
    load_certificate,
    load_certificate_issuer_privatekey,
    select_message_digest,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    get_relative_time_option,
)

if t.TYPE_CHECKING:
    import datetime  # pragma: no cover

    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.module_utils._argspec import (  # pragma: no cover
        ArgumentSpec,
    )


try:
    import cryptography
    from cryptography import x509
    from cryptography.hazmat.primitives.serialization import Encoding
except ImportError:
    pass


class OwnCACertificateBackendCryptography(CertificateBackend):
    def __init__(self, *, module: AnsibleModule) -> None:
        super().__init__(module=module)

        self.create_subject_key_identifier: t.Literal[
            "create_if_not_provided", "always_create", "never_create"
        ] = module.params["ownca_create_subject_key_identifier"]
        self.create_authority_key_identifier: bool = module.params[
            "ownca_create_authority_key_identifier"
        ]
        self.not_before = get_relative_time_option(
            module.params["ownca_not_before"],
            input_name="ownca_not_before",
            with_timezone=CRYPTOGRAPHY_TIMEZONE,
        )
        self.not_after = get_relative_time_option(
            module.params["ownca_not_after"],
            input_name="ownca_not_after",
            with_timezone=CRYPTOGRAPHY_TIMEZONE,
        )
        self.digest = select_message_digest(module.params["ownca_digest"])
        self.serial_number = x509.random_serial_number()
        self.ca_cert_path: str | None = module.params["ownca_path"]
        ca_cert_content: str | None = module.params["ownca_content"]
        if ca_cert_content is not None:
            self.ca_cert_content: bytes | None = ca_cert_content.encode("utf-8")
        else:
            self.ca_cert_content = None
        self.ca_privatekey_path: str | None = module.params["ownca_privatekey_path"]
        ca_privatekey_content: str | None = module.params["ownca_privatekey_content"]
        if ca_privatekey_content is not None:
            self.ca_privatekey_content: bytes | None = ca_privatekey_content.encode(
                "utf-8"
            )
        else:
            self.ca_privatekey_content = None
        self.ca_privatekey_passphrase: str | None = module.params[
            "ownca_privatekey_passphrase"
        ]

        if self.csr_content is None:
            if self.csr_path is None:
                raise CertificateError(
                    "csr_path or csr_content is required for ownca provider"
                )
            if not os.path.exists(self.csr_path):
                raise CertificateError(
                    f"The certificate signing request file {self.csr_path} does not exist"
                )
        if self.ca_cert_path is not None and not os.path.exists(self.ca_cert_path):
            raise CertificateError(
                f"The CA certificate file {self.ca_cert_path} does not exist"
            )
        if self.ca_privatekey_path is not None and not os.path.exists(
            self.ca_privatekey_path
        ):
            raise CertificateError(
                f"The CA private key file {self.ca_privatekey_path} does not exist"
            )

        self._ensure_csr_loaded()
        self.ca_cert = load_certificate(
            path=self.ca_cert_path,
            content=self.ca_cert_content,
        )
        if not is_potential_certificate_issuer_public_key(self.ca_cert.public_key()):
            raise CertificateError(
                "CA certificate's public key cannot be used to sign certificates"
            )
        try:
            self.ca_private_key = load_certificate_issuer_privatekey(
                path=self.ca_privatekey_path,
                content=self.ca_privatekey_content,
                passphrase=self.ca_privatekey_passphrase,
            )
        except OpenSSLBadPassphraseError as exc:
            module.fail_json(msg=str(exc))

        if not cryptography_compare_public_keys(
            self.ca_cert.public_key(), self.ca_private_key.public_key()
        ):
            raise CertificateError(
                "The CA private key does not belong to the CA certificate"
            )

        if cryptography_key_needs_digest_for_signing(self.ca_private_key):
            if self.digest is None:
                raise CertificateError(
                    f"The digest {module.params['ownca_digest']} is not supported with the cryptography backend"
                )
        else:
            self.digest = None

    def generate_certificate(self) -> None:
        """(Re-)Generate certificate."""
        if self.csr is None:
            raise AssertionError(
                "Contract violation: csr has not been populated"
            )  # pragma: no cover
        cert_builder = x509.CertificateBuilder()
        cert_builder = cert_builder.subject_name(self.csr.subject)
        cert_builder = cert_builder.issuer_name(self.ca_cert.subject)
        cert_builder = cert_builder.serial_number(self.serial_number)
        cert_builder = set_not_valid_before(cert_builder, self.not_before)
        cert_builder = set_not_valid_after(cert_builder, self.not_after)
        cert_builder = cert_builder.public_key(self.csr.public_key())
        has_ski = False
        for extension in self.csr.extensions:
            if isinstance(extension.value, x509.SubjectKeyIdentifier):
                if self.create_subject_key_identifier == "always_create":
                    continue
                has_ski = True
            if self.create_authority_key_identifier and isinstance(
                extension.value, x509.AuthorityKeyIdentifier
            ):
                continue
            cert_builder = cert_builder.add_extension(
                extension.value, critical=extension.critical
            )
        if not has_ski and self.create_subject_key_identifier != "never_create":
            cert_builder = cert_builder.add_extension(
                x509.SubjectKeyIdentifier.from_public_key(self.csr.public_key()),
                critical=False,
            )
        if self.create_authority_key_identifier:
            try:
                ext = self.ca_cert.extensions.get_extension_for_class(
                    x509.SubjectKeyIdentifier
                )
                cert_builder = cert_builder.add_extension(
                    (
                        x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
                            ext.value
                        )
                    ),
                    critical=False,
                )
            except cryptography.x509.ExtensionNotFound:
                public_key = self.ca_cert.public_key()
                assert is_potential_certificate_issuer_public_key(public_key)
                cert_builder = cert_builder.add_extension(
                    x509.AuthorityKeyIdentifier.from_issuer_public_key(public_key),
                    critical=False,
                )

        certificate = cert_builder.sign(
            private_key=self.ca_private_key,
            algorithm=self.digest,
        )

        self.cert = certificate

    def get_certificate_data(self) -> bytes:
        """Return bytes for self.cert."""
        if self.cert is None:
            raise AssertionError(
                "Contract violation: cert has not been populated"
            )  # pragma: no cover
        return self.cert.public_bytes(Encoding.PEM)

    def needs_regeneration(
        self,
        *,
        not_before: datetime.datetime | None = None,
        not_after: datetime.datetime | None = None,
    ) -> bool:
        if super().needs_regeneration(
            not_before=self.not_before, not_after=self.not_after
        ):
            return True

        self._ensure_existing_certificate_loaded()
        assert self.existing_certificate is not None

        # Check whether certificate is signed by CA certificate
        if not cryptography_verify_certificate_signature(
            certificate=self.existing_certificate,
            signer_public_key=self.ca_cert.public_key(),
        ):
            return True

        # Check subject
        if self.ca_cert.subject != self.existing_certificate.issuer:
            return True

        # Check AuthorityKeyIdentifier
        if self.create_authority_key_identifier:
            try:
                ext_ski = self.ca_cert.extensions.get_extension_for_class(
                    x509.SubjectKeyIdentifier
                )
                expected_ext = (
                    x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
                        ext_ski.value
                    )
                )
            except cryptography.x509.ExtensionNotFound:
                public_key = self.ca_cert.public_key()
                assert is_potential_certificate_issuer_public_key(public_key)
                expected_ext = x509.AuthorityKeyIdentifier.from_issuer_public_key(
                    public_key
                )

            try:
                ext_aki = self.existing_certificate.extensions.get_extension_for_class(
                    x509.AuthorityKeyIdentifier
                )
                if ext_aki.value != expected_ext:
                    return True
            except cryptography.x509.ExtensionNotFound:
                return True

        return False

    def dump(self, *, include_certificate: bool) -> dict[str, t.Any]:
        result = super().dump(include_certificate=include_certificate)
        result.update(
            {
                "ca_cert": self.ca_cert_path,
                "ca_privatekey": self.ca_privatekey_path,
            }
        )

        if self.module.check_mode:
            result.update(
                {
                    "notBefore": self.not_before.strftime("%Y%m%d%H%M%SZ"),
                    "notAfter": self.not_after.strftime("%Y%m%d%H%M%SZ"),
                    "serial_number": self.serial_number,
                }
            )
        else:
            if self.cert is None:
                self.cert = self.existing_certificate
            assert self.cert is not None
            result.update(
                {
                    "notBefore": get_not_valid_before(self.cert).strftime(
                        "%Y%m%d%H%M%SZ"
                    ),
                    "notAfter": get_not_valid_after(self.cert).strftime(
                        "%Y%m%d%H%M%SZ"
                    ),
                    "serial_number": self.cert.serial_number,
                }
            )

        return result


def generate_serial_number() -> int:
    """Generate a serial number for a certificate"""
    while True:
        result = randrange(0, 1 << 160)
        if result >= 1000:
            return result


class OwnCACertificateProvider(CertificateProvider):
    def validate_module_args(self, module: AnsibleModule) -> None:
        if (
            module.params["ownca_path"] is None
            and module.params["ownca_content"] is None
        ):
            module.fail_json(
                msg="One of ownca_path and ownca_content must be specified for the ownca provider."
            )
        if (
            module.params["ownca_privatekey_path"] is None
            and module.params["ownca_privatekey_content"] is None
        ):
            module.fail_json(
                msg="One of ownca_privatekey_path and ownca_privatekey_content must be specified for the ownca provider."
            )

    def create_backend(
        self, module: AnsibleModule
    ) -> OwnCACertificateBackendCryptography:
        return OwnCACertificateBackendCryptography(module=module)


def add_ownca_provider_to_argument_spec(argument_spec: ArgumentSpec) -> None:
    argument_spec.argument_spec["provider"]["choices"].append("ownca")  # type: ignore
    argument_spec.argument_spec.update(
        {
            "ownca_path": {"type": "path"},
            "ownca_content": {"type": "str"},
            "ownca_privatekey_path": {"type": "path"},
            "ownca_privatekey_content": {"type": "str", "no_log": True},
            "ownca_privatekey_passphrase": {"type": "str", "no_log": True},
            "ownca_digest": {"type": "str", "default": "sha256"},
            "ownca_version": {"type": "int", "default": 3, "choices": [3]},  # not used
            "ownca_not_before": {"type": "str", "default": "+0s"},
            "ownca_not_after": {"type": "str", "default": "+3650d"},
            "ownca_create_subject_key_identifier": {
                "type": "str",
                "default": "create_if_not_provided",
                "choices": ["create_if_not_provided", "always_create", "never_create"],
            },
            "ownca_create_authority_key_identifier": {"type": "bool", "default": True},
        }
    )
    argument_spec.mutually_exclusive.extend(
        [
            ["ownca_path", "ownca_content"],
            ["ownca_privatekey_path", "ownca_privatekey_content"],
        ]
    )


__all__ = (
    "OwnCACertificateBackendCryptography",
    "OwnCACertificateProvider",
    "add_ownca_provider_to_argument_spec",
)



# Copyright (c) 2019-2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import os.path
import selectors
import socket as pysocket
import struct
import typing as t

from ansible_collections.community.docker.plugins.module_utils._api.utils import (
    socket as docker_socket,
)
from ansible_collections.community.docker.plugins.module_utils._socket_helper import (
    make_unblocking,
    shutdown_writing,
    write_to_socket,
)

if t.TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType

    from ansible.module_utils.basic import AnsibleModule

    from ansible_collections.community.docker.plugins.module_utils._socket_helper import (
        SocketLike,
    )


PARAMIKO_POLL_TIMEOUT = 0.01  # 10 milliseconds


def _empty_writer(msg: str) -> None:
    pass


class DockerSocketHandlerBase:
    def __init__(
        self, sock: SocketLike, log: Callable[[str], None] | None = None
    ) -> None:
        make_unblocking(sock)

        self._log = log or _empty_writer
        self._paramiko_read_workaround = hasattr(
            sock, "send_ready"
        ) and "paramiko" in str(type(sock))

        self._sock = sock
        self._block_done_callback: Callable[[int, bytes], None] | None = None
        self._block_buffer: list[tuple[int, bytes]] = []
        self._eof = False
        self._read_buffer = b""
        self._write_buffer = b""
        self._end_of_writing = False

        self._current_stream: int | None = None
        self._current_missing = 0
        self._current_buffer = b""

        self._selector = selectors.DefaultSelector()
        self._selector.register(self._sock, selectors.EVENT_READ)

    def __enter__(self) -> t.Self:
        return self

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self._selector.close()

    def set_block_done_callback(
        self, block_done_callback: Callable[[int, bytes], None]
    ) -> None:
        self._block_done_callback = block_done_callback
        if self._block_done_callback is not None:
            while self._block_buffer:
                elt = self._block_buffer.pop(0)
                self._block_done_callback(*elt)

    def _add_block(self, stream_id: int, data: bytes) -> None:
        if self._block_done_callback is not None:
            self._block_done_callback(stream_id, data)
        else:
            self._block_buffer.append((stream_id, data))

    def _read(self) -> None:
        if self._eof:
            return
        data: bytes | None
        if hasattr(self._sock, "recv"):
            try:
                data = self._sock.recv(262144)
            except Exception as e:  # pylint: disable=broad-exception-caught
                # After calling self._sock.shutdown(), OpenSSL's/urllib3's
                # WrappedSocket seems to eventually raise ZeroReturnError in
                # case of EOF
                if "OpenSSL.SSL.ZeroReturnError" in str(type(e)):
                    self._eof = True
                    return
                raise
        elif isinstance(self._sock, pysocket.SocketIO):  # type: ignore[unreachable]
            data = self._sock.read()  # type: ignore
        else:
            data = os.read(self._sock.fileno())  # type: ignore  # TODO does this really work?!
        if data is None:
            # no data available
            return  # type: ignore[unreachable]
        self._log(f"read {len(data)} bytes")
        if len(data) == 0:
            # Stream EOF
            self._eof = True
            return
        self._read_buffer += data
        while len(self._read_buffer) > 0:
            if self._current_missing > 0:
                n = min(len(self._read_buffer), self._current_missing)
                self._current_buffer += self._read_buffer[:n]
                self._read_buffer = self._read_buffer[n:]
                self._current_missing -= n
                if self._current_missing == 0:
                    assert self._current_stream is not None
                    self._add_block(self._current_stream, self._current_buffer)
                    self._current_buffer = b""
            if len(self._read_buffer) < 8:
                break
            self._current_stream, self._current_missing = struct.unpack(
                ">BxxxL", self._read_buffer[:8]
            )
            self._read_buffer = self._read_buffer[8:]
            if self._current_missing < 0:
                # Stream EOF (as reported by docker daemon)
                self._eof = True
                break

    def _handle_end_of_writing(self) -> None:
        if self._end_of_writing and len(self._write_buffer) == 0:
            self._end_of_writing = False
            self._log("Shutting socket down for writing")
            shutdown_writing(self._sock, self._log)

    def _write(self) -> None:
        if len(self._write_buffer) > 0:
            written = write_to_socket(self._sock, self._write_buffer)
            self._write_buffer = self._write_buffer[written:]
            self._log(f"wrote {written} bytes, {len(self._write_buffer)} are left")
            if len(self._write_buffer) > 0:
                self._selector.modify(
                    self._sock, selectors.EVENT_READ | selectors.EVENT_WRITE
                )
            else:
                self._selector.modify(self._sock, selectors.EVENT_READ)
            self._handle_end_of_writing()

    def select(
        self, timeout: int | float | None = None, _internal_recursion: bool = False
    ) -> bool:
        if (
            not _internal_recursion
            and self._paramiko_read_workaround
            and len(self._write_buffer) > 0
        ):
            # When the SSH transport is used, Docker SDK for Python internally uses Paramiko, whose
            # Channel object supports select(), but only for reading
            # (https://github.com/paramiko/paramiko/issues/695).
            if self._sock.send_ready():  # type: ignore
                self._write()
                return True
            while timeout is None or timeout > PARAMIKO_POLL_TIMEOUT:
                result = int(
                    self.select(PARAMIKO_POLL_TIMEOUT, _internal_recursion=True)
                )
                if self._sock.send_ready():  # type: ignore
                    self._read()
                    result += 1
                if result > 0:
                    return True
                if timeout is not None:
                    timeout -= PARAMIKO_POLL_TIMEOUT
        self._log(f"select... ({timeout})")
        events = self._selector.select(timeout)
        for key, event in events:
            if key.fileobj == self._sock:
                ev_read = event & selectors.EVENT_READ != 0
                ev_write = event & selectors.EVENT_WRITE != 0
                self._log(f"select event read:{ev_read} write:{ev_write}")
                if event & selectors.EVENT_READ != 0:
                    self._read()
                if event & selectors.EVENT_WRITE != 0:
                    self._write()
        result = len(events)
        if self._paramiko_read_workaround and len(self._write_buffer) > 0 and self._sock.send_ready():  # type: ignore
            self._write()
            result += 1
        return result > 0

    def is_eof(self) -> bool:
        return self._eof

    def end_of_writing(self) -> None:
        self._end_of_writing = True
        self._handle_end_of_writing()

    def consume(self) -> tuple[bytes, bytes]:
        stdout = []
        stderr = []

        def append_block(stream_id: int, data: bytes) -> None:
            if stream_id == docker_socket.STDOUT:
                stdout.append(data)
            elif stream_id == docker_socket.STDERR:
                stderr.append(data)
            else:
                raise ValueError(f"{stream_id} is not a valid stream ID")

        self.end_of_writing()

        self.set_block_done_callback(append_block)
        while not self._eof:
            self.select()
        return b"".join(stdout), b"".join(stderr)

    def write(self, str_to_write: bytes) -> None:
        self._write_buffer += str_to_write
        if len(self._write_buffer) == len(str_to_write):
            self._write()


class DockerSocketHandlerModule(DockerSocketHandlerBase):
    def __init__(self, sock: SocketLike, module: AnsibleModule) -> None:
        super().__init__(sock, module.debug)
